"""
Term Repository tests.

Tests lute.term.model.Term *domain* objects being saved
and retrieved from DB.
"""

from datetime import datetime
import pytest

from lute.models.term import Term as DBTerm, TermTag
from lute.db import db
from lute.term.model import Term, Repository
from tests.dbasserts import assert_sql_result, assert_record_count_equals
from tests.utils import add_terms, make_text


@pytest.fixture(name="repo")
def fixture_repo():
    return Repository(db)


@pytest.fixture(name="hello_term")
def fixture_hello_term(english):
    """
    Term business object with some defaults,
    no tags or parents.
    """
    t = Term()
    t.language_id = english.id
    t.text = "HELLO"
    t.translation = "greeting"
    t.current_image = "hello.png"
    t.flash_message = "hello flash"
    return t


def test_save_new(app_context, hello_term, repo):
    """
    Saving a simple Term object loads the database.
    """
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    repo.add(hello_term)
    assert_sql_result(sql, [], "Still empty")

    repo.commit()
    assert_sql_result(sql, ["HELLO; hello; 1"], "Saved")

    term = repo.find(hello_term.language_id, hello_term.text)
    assert term.text == hello_term.text


def test_save_new_multiword(app_context, hello_term, repo):
    """
    Zero-width strings are added between parsed tokens.
    """
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    hello_term.text = "HELLO THERE"
    repo.add(hello_term)
    assert_sql_result(sql, [], "Still empty")

    repo.commit()
    # Assert replaces zws with '/'
    assert_sql_result(sql, ["HELLO/ /THERE; hello/ /there; 3"], "Saved")


def test_save_updates_existing(english, app_context, hello_term, repo):
    """
    Saving Term updates an existing term if the text and lang
    matches.
    """

    term = DBTerm(english, "HELLO")
    db.session.add(term)
    db.session.commit()
    sql = "select WoID, WoText, WoStatus from words"
    assert_sql_result(sql, [f"{term.id}; HELLO; 1"], "have term")

    hello_term.language_id = english.id
    hello_term.text = "hello"
    hello_term.status = 5

    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [f"{term.id}; hello; 5"], "have term, status changed")


def test_save_existing_replaces_tags(english, app_context, repo):
    """
    If the db has a term with tags, and the business term changes
    those, the existing tags are replaced.
    """
    dbterm = DBTerm(english, "HELLO")
    dbterm.add_term_tag(TermTag("a"))
    dbterm.add_term_tag(TermTag("b"))
    db.session.add(dbterm)
    db.session.commit()
    sql = f"""select TgText from tags
    inner join wordtags on WtTgID = TgID
    where WtWoID = {dbterm.id}"""
    assert_sql_result(sql, ["a", "b"], "have term tags")

    term = repo.load(dbterm.id)
    assert term.term_tags == ["a", "b"]

    term.term_tags = ["a", "c"]
    repo.add(term)
    repo.commit()
    assert_sql_result(sql, ["a", "c"], "term tags changed")

    term.term_tags = []
    repo.add(term)
    repo.commit()
    assert_sql_result(sql, [], "term tags removed")

    sql = "select TgText from tags order by TgText"
    assert_sql_result(sql, ["a", "b", "c"], "Source tags still exist")


def test_save_uses_existing_TermTags(app_context, repo, hello_term):
    "Don't create new TermTag records if they already exist."
    db.session.add(TermTag("a"))
    db.session.commit()

    sql = """select TgID, TgText, WoText
    from tags
    left join wordtags on WtTgID = TgID
    left join words on WoID = WtWoID
    order by TgText"""
    assert_sql_result(sql, ["1; a; None"], "a tag exists")

    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["1; a; HELLO", "2; b; HELLO"], "a used, b created")


def test_save_with_no_flash_message(app_context, repo, hello_term):
    "Saving with flash = None removes the flash record."
    hello_term.flash_message = "hi there"
    repo.add(hello_term)
    repo.commit()
    sql = "select WfMessage from wordflashmessages"
    assert_sql_result(sql, ["hi there"], "record exists")

    hello_term.flash_message = None
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "removed")


## Deletes.


def test_delete(app_context, repo, hello_term):
    "Removes record from db."
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC, WoStatus from words"
    assert_sql_result(sql, ["hello; 1"], "record exists")

    repo.delete(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "deleted")


def test_delete_leaves_parent(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words"
    assert_sql_result(sql, ["hello", "parent"], "records exists")

    repo.delete(hello_term)
    repo.commit()
    assert_sql_result(sql, ["parent"], "parent stays")


## Saving and images.


def test_save_term_image_saved_and_replaced(app_context, repo, hello_term):
    "Saving puts record in image table."
    sql = "select WiSource from wordimages"
    assert_sql_result(sql, [], "nothing present")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello.png"], "image saved")


def test_save_term_image_set_to_blank_removes_record(app_context, repo, hello_term):
    "Saving puts record in image table."
    sql = "select WiSource from wordimages"
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello.png"], "image saved")

    hello_term.current_image = ""
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "record removed")

    hello_term.current_image = "new.png"
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["new.png"], "image saved")

    hello_term.current_image = None
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "record removed again")


## Saving and parents.


def test_save_with_new_parent(app_context, repo, hello_term):
    """
    Given a Term with parents = [ 'newparent' ],
    new parent DBTerm is created, and is assigned translation and image and tag.
    """
    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result("select WoText from words", ["HELLO", "parent"], "parent created")

    parent = repo.find(hello_term.language_id, "parent")
    assert isinstance(parent, Term), "is a Term bus. object"
    assert parent.text == "parent"
    assert parent.term_tags == hello_term.term_tags
    assert parent.term_tags == ["a", "b"]  # just spelling it out.
    assert parent.translation == hello_term.translation
    assert parent.current_image == hello_term.current_image
    assert parent.parents == []


def test_save_remove_parent_breaks_link(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words"
    assert_sql_result(sql, ["hello", "parent"], "records exists")
    assert_record_count_equals("wordparents", 1, "linked")

    hello_term.parents.remove("parent")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello", "parent"], "both stay")
    assert_record_count_equals("wordparents", 0, "link broken")


def test_save_change_parent_breaks_old_link(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, ["hello", "parent"], "records exists")

    sqlparent = """
    select c.WoTextLC, p.WoTextLC
    from wordparents
    inner join words as c on WpWoID = c.WoID
    inner join words as p on WpParentWoID = p.WoID
    """
    assert_sql_result(sqlparent, ["hello; parent"], "records exists")

    hello_term.parents.remove("parent")
    hello_term.parents.append("new")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello", "new", "parent"], "all stay")
    assert_sql_result(sqlparent, ["hello; new"], "changed")


def test_save_term_with_multiple_parents_unsets_sync_status(
    app_context, repo, hello_term
):
    "Ensure DB data is good on save."
    hello_term.parents.append("parent")
    hello_term.sync_status = True
    repo.add(hello_term)
    repo.commit()

    sql = "select WoTextLC, WoSyncStatus from words where WoTextLC = 'hello'"
    assert_sql_result(sql, ["hello; 1"], "link set")

    hello_term.parents.append("parent2")
    hello_term.sync_status = True
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello; 0"], "link UN-set")


def test_cant_set_term_as_its_own_parent(app_context, repo, hello_term):
    """
    Would create obvious circular ref
    """
    hello_term.parents = [hello_term.text]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result("select WoText from words", ["HELLO"], "term entered")
    assert_sql_result("select * from wordparents", [], "no records")


@pytest.fixture(name="_existing_parent")
def fixture_existing_parent(app_context, english):
    "Create an existing parent."
    term = DBTerm(english, "parent")
    db.session.add(term)
    db.session.commit()
    return term


def test_save_with_existing_parent_creates_link(
    app_context, repo, hello_term, _existing_parent
):
    """
    new parent link is created, and is assigned translation and image and tag.
    """

    sql = "select WoID, WoText from words"
    assert_sql_result(sql, ["1; parent"], "initial state")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    dbterm = repo.add(hello_term)
    repo.commit()

    assert_sql_result(sql, ["1; parent", "2; HELLO"], "new record added")
    assert len(dbterm.parents) == 1
    assert dbterm.parents[0].text == "parent"


def test_save_existing_parent_gets_translation_if_missing(
    app_context, repo, hello_term, _existing_parent
):
    """
    Translation and image applied if missing.
    """

    sql = "select WoID, WoText, WoTranslation from words"
    assert_sql_result(sql, ["1; parent; None"], "initial state")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    assert_sql_result(
        sql, ["1; parent; greeting", "2; HELLO; greeting"], "transl saved"
    )


## Find tests.


def test_load(empty_db, english, repo):
    "Smoke test."
    t = DBTerm(english, "Hello")
    t.set_current_image("hello.png")
    t.add_term_tag(TermTag("a"))
    t.add_term_tag(TermTag("b"))
    db.session.add(t)
    db.session.commit()
    assert t.id is not None, "ID assigned"

    term = repo.load(t.id)
    assert term.id == t.id
    assert term.language.id == t.language.id, "lang object set"
    assert term.language_id == english.id
    assert term.text == "Hello"
    assert term.original_text == "Hello"
    assert term.current_image == "hello.png"
    assert len(term.parents) == 0
    assert term.term_tags == ["a", "b"]


def test_load_throws_if_bad_id(app_context, repo):
    "If app says term id = 7 exists, and it doesn't, that's a problem."
    with pytest.raises(ValueError):
        repo.load(9876)


def test_find_is_found(spanish, app_context, repo):
    "Find by text finds regardless of case."
    add_terms(spanish, ["PARENT"])
    cases = ["PARENT", "parent", "pAReNt"]
    for c in cases:
        p = repo.find(spanish.id, c)
        assert p is not None, f"parent found for case {c}"
        assert p.text == "PARENT", f"parent found for case {c}"


def test_find_not_found_returns_none(spanish, repo):
    "No match = none."
    p = repo.find(spanish.id, "unknown_term")
    assert p is None, "nothing found"


def test_find_only_looks_in_specified_language(spanish, english, repo):
    "Only search language."
    add_terms(english, ["hola"])
    p = repo.find(spanish.id, "hola")
    assert p is None, "english terms not checked"


def test_find_existing_multi_word(spanish, repo):
    "Domain objects don't have zero-width strings in them."
    add_terms(spanish, ["una bebida"])
    zws = "\u200B"
    t = repo.find(spanish.id, f"una{zws} {zws}bebida")
    assert t.id > 0
    assert t.text == "una bebida"

    t = repo.find(spanish.id, "una bebida")
    assert t.id > 0
    assert t.text == "una bebida"


## Find or new tests.


def test_find_or_new_existing_word(spanish, repo):
    "Existing term is found."
    add_terms(spanish, ["BEBIDA"])
    t = repo.find_or_new(spanish.id, "bebida")
    assert t.id > 0, "exists"
    assert t.text == "BEBIDA"
    assert t.language.id == spanish.id, "lang object set"


def test_find_or_new_non_existing(spanish, repo):
    "Returns new term."
    t = repo.find_or_new(spanish.id, "TENGO")
    assert t.id is None
    assert t.text == "TENGO"
    assert t.language.id == spanish.id, "lang object set"


def test_find_or_new_existing_multi_word(spanish, repo):
    "Spaces etc handled correctly."
    add_terms(spanish, ["una bebida"])
    zws = "\u200B"
    t = repo.find_or_new(spanish.id, f"una{zws} {zws}bebida")
    assert t.id > 0
    assert t.text == "una bebida"

    t = repo.find_or_new(spanish.id, "una bebida")
    assert t.id > 0
    assert t.text == "una bebida"


def test_find_or_new_new_multi_word(spanish, repo):
    "ZWS added correctly."
    zws = "\u200B"
    t = repo.find_or_new(spanish.id, f"una{zws} {zws}bebida")
    assert t.id is None
    assert t.text == f"una{zws} {zws}bebida"

    t = repo.find_or_new(spanish.id, "una bebida")
    assert t.id is None
    assert t.text == "una bebida"


## Matches tests.


@pytest.fixture(name="_multiple_terms")
def fixture_multiple(english, spanish, app_context):
    "Create multiple terms for find_matches tests."
    add_terms(english, ["parent"])
    add_terms(spanish, ["parent", "pare", "gato", "tengo uno"])


@pytest.mark.find_match
def test_find_matches_only_returns_language_matches(spanish, repo, _multiple_terms):
    "Searches match the start of string."
    for c in ["PARE", "pare", "PAR"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 2, c
        assert matches[0].text == "pare"
        assert matches[0].language.id == spanish.id, "lang object set"


@pytest.mark.find_match
def test_find_matches_returns_empty_if_no_match_or_empty_string(
    spanish, repo, _multiple_terms
):
    "Empty if no match."
    for c in ["", "x"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 0, c


@pytest.mark.find_match
def test_find_matches_multiword_respects_zws(spanish, repo, _multiple_terms):
    "zws are removed from the business objects."
    zws = "\u200B"
    for c in [f"tengo{zws} {zws}uno", "tengo uno", "tengo"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 1, f'have match for case "{c}"'
        assert matches[0].text == "tengo uno"


def assert_find_matches_returns(term_repo, spanish, s, expected):
    "Helper, assert returns expected content."
    matches = term_repo.find_matches(spanish.id, s)
    actual = ", ".join([t.text for t in matches])
    assert ", ".join(expected) == actual


def test_find_matches_starting_matches_sort_to_top(spanish, repo):
    "All matches return, but words that start with the same thing sort to top."
    add_terms(spanish, ["tener", "contener", "sostener", "ten", "contengo", "xxx"])
    assert_find_matches_returns(
        repo, spanish, "ten", ["ten", "tener", "contener", "contengo", "sostener"]
    )
    assert_find_matches_returns(repo, spanish, "ene", ["contener", "sostener", "tener"])
    assert_find_matches_returns(repo, spanish, "x", ["xxx"])


@pytest.mark.find_match
def test_findLikeSpecification_initial_check(spanish, repo):
    "Searches match any part of string."
    add_terms(spanish, ["abc", "abcd", "bcd"])

    assert_find_matches_returns(repo, spanish, "ab", ["abc", "abcd"])
    assert_find_matches_returns(repo, spanish, "abcd", ["abcd"])
    assert_find_matches_returns(repo, spanish, "bcd", ["bcd", "abcd"])
    assert_find_matches_returns(repo, spanish, "yy", [])


@pytest.mark.find_match
def test_find_like_specification_terms_with_children_go_to_top(spanish, repo):
    """
    Parents go to the top, then the rest ... but exact matches trumps parent.
    """
    terms = [("abc", "abcParent"), ("axy", "axyParent")]
    for term, parent in terms:
        t = Term()
        t.language = spanish
        t.language_id = spanish.id
        t.text = term
        t.parents.append(parent)
        repo.add(t)
    repo.commit()

    assert_find_matches_returns(
        repo, spanish, "a", ["abcParent", "axyParent", "abc", "axy"]
    )
    assert_find_matches_returns(repo, spanish, "ab", ["abcParent", "abc"])
    assert_find_matches_returns(repo, spanish, "abc", ["abc", "abcParent"])


## Sentences


def full_refs_to_string(refs):
    "Convert refs to strings for simpler testing."

    def to_string(r):
        return f"{r.title}, {r.sentence}"

    def refs_to_string(refs_array):
        ret = [to_string(r) for r in refs_array]
        ret.sort()
        return ret

    def parent_refs_to_string(prefs):
        ret = []
        for p in prefs:
            ret.append({"term": p["term"], "refs": refs_to_string(p["refs"])})
        return ret

    return {
        "term": refs_to_string(refs["term"]),
        "children": refs_to_string(refs["children"]),
        "parents": parent_refs_to_string(refs["parents"]),
    }


@pytest.mark.sentences
def test_get_all_references(spanish, repo):
    "Check references with parents and children."
    text = make_text(
        "hola", "Tengo un gato.  Ella tiene un perro.  No quiero tener nada.", spanish
    )
    archtext = make_text("luego", "Tengo un coche.", spanish)
    archtext.book.archived = True
    db.session.add(archtext.book)

    for t in [text, archtext]:
        t.read_date = datetime.now()
        db.session.add(t)

    db.session.commit()

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    tengo.parents = ["tener"]
    repo.add(tengo)

    tiene = Term()
    tiene.language_id = spanish.id
    tiene.text = "tiene"
    tiene.parents = ["tener"]
    repo.add(tiene)

    repo.commit()

    refs = repo.find_references(tengo)
    assert full_refs_to_string(refs) == {
        "term": [
            "hola (1/1), <b>Tengo</b> un gato.",
            "luego (1/1), <b>Tengo</b> un coche.",
        ],
        "children": [],
        "parents": [
            {
                "term": "tener",
                "refs": [
                    "hola (1/1), Ella <b>tiene</b> un perro.",
                    "hola (1/1), No quiero <b>tener</b> nada.",
                ],
            }
        ],
    }, "term tengo"

    tener = repo.find(spanish.id, "tener")
    refs = repo.find_references(tener)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), No quiero <b>tener</b> nada."],
        "children": [
            "hola (1/1), <b>Tengo</b> un gato.",
            "hola (1/1), Ella <b>tiene</b> un perro.",
            "luego (1/1), <b>Tengo</b> un coche.",
        ],
        "parents": [],
    }, "term tener"


@pytest.mark.sentences
def test_multiword_reference(spanish, repo):
    "Ensure zws-delimiters are respected."
    text = make_text("hola", "Yo tengo un gato.", spanish)
    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()

    t = Term()
    t.language_id = spanish.id
    t.text = "TENGO UN"
    repo.add(t)
    repo.commit()

    refs = repo.find_references(t)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), Yo <b>tengo un</b> gato."],
        "children": [],
        "parents": [],
    }, "term tengo"


@pytest.mark.sentences
def test_get_references_only_includes_read_texts(spanish, repo):
    "Like it says above."
    text = make_text("hola", "Tengo un gato.  No tengo un perro.", spanish)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    repo.add(tengo)
    repo.commit()

    refs = repo.find_references(tengo)
    keys = refs.keys()
    for k in keys:
        assert len(refs[k]) == 0, f"{k}, no matches for unread texts"

    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()

    refs = repo.find_references(tengo)
    assert len(refs["term"]) == 2, "have refs once text is read"


@pytest.mark.sentences
def test_get_references_only_includes_refs_in_same_language(spanish, english, repo):
    "Like it says above."
    text1 = make_text("hola", "Tengo un gato.  No tengo un perro.", spanish)
    text2 = make_text("hola", "Tengo in english.", english)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    repo.add(tengo)
    repo.commit()

    text1.read_date = datetime.now()
    text2.read_date = datetime.now()
    db.session.add(text1)
    db.session.add(text2)
    db.session.commit()

    refs = repo.find_references(tengo)
    assert len(refs["term"]) == 2, "only have 2 refs (spanish)"
    sentences = [r.sentence for r in refs["term"]]
    assert "<b>Tengo</b> un gato." in sentences
    assert "No <b>tengo</b> un perro." in sentences
    assert "<b>Tengo</b> in english." not in sentences


@pytest.mark.sentences
def test_get_references_new_term(spanish, repo):
    "Check references with parents and children."
    text = make_text("hola", "Tengo un gato.", spanish)
    text.read_date = datetime.now()
    db.session.add(text)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"

    refs = repo.find_references(tengo)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), <b>Tengo</b> un gato."],
        "children": [],
        "parents": [],
    }, "term tengo"



"""
Term tests.
"""

import pytest
from sqlalchemy import text
from lute.models.term import Term, TermImage
from lute.db import db
from tests.dbasserts import assert_record_count_equals, assert_sql_result


def test_cruft_stripped_on_set_word(spanish):
    "Extra spaces are stripped, because they cause parsing/matching problems."
    cases = [
        ("hola", "hola", "hola"),
        ("    hola    ", "hola", "hola"),
        # This case should never occur:
        # tabs are stripped out of text, and returns mark different sentences.
        # ('   hola\tGATO\nok', 'hola GATO ok', 'hola gato ok'),
    ]

    for s, expected_text, expected_text_lc in cases:
        term = Term(spanish, s)
        assert term.text == expected_text
        assert term.text_lc == expected_text_lc


def test_token_count(english):
    "Various scenarios."
    cases = [
        ("hola", 1),
        ("    hola    ", 1),
        ("  hola  gato", 3),
        ("HOLA hay\tgato  ", 5),
        ("  the CAT's pyjamas  ", 7),
        # This only has 9 tokens, because the "'" is included with
        # the following space ("' ").
        ("A big CHUNK O' stuff", 9),
        ("YOU'RE", 3),
        ("...", 1),  # should never happen :-)
    ]

    for s, expected_token_count in cases:
        term = Term(english, s)
        assert term.token_count == expected_token_count


def test_term_left_as_is_if_its_an_exception(spanish):
    "Ensure regex match works, upper and lowercase."
    spanish.exceptions_split_sentences = "EE.UU."

    term = Term(spanish, "EE.UU.")
    assert term.token_count == 1
    assert term.text == "EE.UU."

    term = Term(spanish, "ee.uu.")
    assert term.token_count == 1
    assert term.text == "ee.uu."


def test_cannot_add_self_as_own_parent(spanish):
    "Avoid circular references."
    t = Term(spanish, "gato")
    t.add_parent(t)
    assert len(t.parents) == 0

    otherterm = Term(spanish, "gato")
    t.add_parent(otherterm)
    assert len(t.parents) == 0, "different object still not added"


def test_find_by_spec(app_context, spanish, english):
    """
    Can find by spec, matches on language and text.
    """
    t = Term(spanish, "gato")
    db.session.add(t)
    db.session.commit()

    spec = Term(spanish, "GATO")
    found = Term.find_by_spec(spec)
    assert found.id == t.id, "term found by matching spec"

    spec = Term(english, "GATO")
    found = Term.find_by_spec(spec)
    assert found is None, "not found in different language"

    spec = Term(spanish, "gatito")
    found = Term.find_by_spec(spec)
    assert found is None, "not found with different text"


# WoStatusChanged checks.
#
# Saving a term changes WoStatusChanged date, via a db trigger.


@pytest.fixture(name="_saved_term")
def fixture_saved_term(app_context, english):
    "Saved term."
    t = Term(english, "hello")
    db.session.add(t)
    db.session.commit()
    return t


def _get_field_value(_saved_term):
    "Get the updated date, and the timestamp."
    sql = text(
        """
        SELECT
          WoStatusChanged,
          ROUND((JULIANDAY(datetime('now')) - JULIANDAY(WoStatusChanged)) * 86400) as diffsecs
        FROM words
        WHERE WoID = :term_id
    """
    )

    result = db.session.execute(sql, {"term_id": _saved_term.id}).fetchone()
    diff = int(result[1])
    return [result[0], diff]


def _assert_updated(_saved_term):
    "Assert the status field was updated."
    val, diff = _get_field_value(_saved_term)
    msg = f"Was updated (set to {val})"
    assert diff <= 100, msg


@pytest.mark.term_status_change
def test_set_on_save_new(app_context, _saved_term):
    "On save of new Term, status changed is creation date."
    _assert_updated(_saved_term)

    sql = "select * from words where WoCreated != WoStatusChanged"
    assert_record_count_equals(sql, 0, "status changed matches created date")


@pytest.mark.term_status_change
def test_update_status_updates_date(app_context, _saved_term):
    "On update, if status has changed, change date."
    db.session.execute(text('update words set WoStatusChanged = "2000-01-01 12:00:00"'))
    db.session.commit()
    sql = 'select * from words where WoStatusChanged = "2000-01-01 12:00:00"'
    assert_record_count_equals(sql, 1, "WoStatusChanged set to test value")

    _saved_term.status = 2
    db.session.add(_saved_term)
    db.session.commit()
    _assert_updated(_saved_term)

    assert_record_count_equals(sql, 0, "updated")


@pytest.mark.term_status_change
def test_saving_with_unchanged_status_leaves_date(app_context, _saved_term):
    "On update, if status has not changed, don't change date."
    db.session.execute(text('update words set WoStatusChanged = "2000-01-01 12:00:00"'))
    db.session.commit()
    _saved_term.status = _saved_term.status
    db.session.add(_saved_term)
    db.session.commit()
    sql = 'select * from words where WoStatusChanged = "2000-01-01 12:00:00"'
    assert_record_count_equals(sql, 1, "WoStatusChanged not updated")


@pytest.mark.term_status_change
def test_update_status_via_sql_updates_date(app_context, _saved_term):
    "On update, if status has changed, change date."
    db.session.execute(text('update words set WoStatusChanged = "2000-01-01 12:00:00"'))
    db.session.commit()
    sql = 'select * from words where WoStatusChanged = "2000-01-01 12:00:00"'
    assert_record_count_equals(sql, 1, "WoStatusChanged set to test value")

    db.session.execute(text("update words set WoStatus = 1"))
    db.session.commit()
    assert_record_count_equals(sql, 1, "same status = same WoStatusChanged")

    db.session.execute(text("update words set WoStatus = 2"))
    db.session.commit()
    assert_record_count_equals(sql, 0, "updated WoStatusChanged")


def test_save_new_image_all_existing_images_replaced(app_context, spanish):
    "All existing terms removed."
    t = Term(spanish, "hola")
    ti1 = TermImage()
    ti1.term = t
    ti1.source = "1.png"
    t.images.append(ti1)

    ti2 = TermImage()
    ti2.term = t
    ti2.source = "2.png"
    t.images.append(ti2)
    db.session.add(t)
    db.session.commit()

    assert_record_count_equals("select * from wordimages", 2, "image count")

    t.set_current_image("3.png")
    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("select * from wordimages", 1, "new image count")
    assert_record_count_equals(
        "select * from wordimages where wisource='3.png'", 1, "new image count"
    )


def test_delete_empty_image_records(app_context, spanish):
    "Check cleanup."
    t = Term(spanish, "hola")
    for s in ["", "   ", "3.png"]:
        ti = TermImage()
        ti.term = t
        ti.source = s
        t.images.append(ti)
    db.session.add(t)
    db.session.commit()

    assert_sql_result("select wisource from wordimages", ["", "   ", "3.png"], "images")

    Term.delete_empty_images()
    assert_sql_result("select wisource from wordimages", ["3.png"], "cleaned images")



"""
Reading helpers.
"""

from lute.models.term import Term, Status
from lute.models.book import Text
from lute.book.stats import mark_stale
from lute.read.render.service import get_paragraphs
from lute.term.model import Repository

from lute.db import db


def set_unknowns_to_known(text: Text):
    """
    Given a text, create new Terms with status Well-Known
    for any new Terms.
    """
    language = text.book.language

    sentences = sum(get_paragraphs(text.text, text.book.language), [])

    tis = []
    for sentence in sentences:
        for ti in sentence.textitems:
            tis.append(ti)

    def is_unknown(ti):
        return (
            ti.is_word == 1
            and (ti.wo_id == 0 or ti.wo_id is None)
            and ti.token_count == 1
        )

    unknowns = list(filter(is_unknown, tis))
    words_lc = [ti.text_lc for ti in unknowns]
    uniques = list(set(words_lc))
    uniques.sort()

    batch_size = 100
    i = 0

    # There is likely a better way to write this using generators and
    # yield.
    for u in uniques:
        candidate = Term(language, u)
        t = Term.find_by_spec(candidate)
        if t is None:
            candidate.status = Status.WELLKNOWN
            db.session.add(candidate)
            i += 1

        if i % batch_size == 0:
            db.session.commit()

    # Commit any remaining.
    db.session.commit()


def bulk_status_update(text: Text, terms_text_array, new_status):
    """
    Given a text and list of terms, update or create new terms
    and set the status.
    """
    language = text.book.language
    repo = Repository(db)
    for term_text in terms_text_array:
        t = repo.find_or_new(language.id, term_text)
        t.status = new_status
        repo.add(t)
    repo.commit()


def start_reading(dbbook, pagenum, db_session):
    "Start reading a page in the book, getting paragraphs."

    text = dbbook.text_at_page(pagenum)
    text.load_sentences()

    mark_stale(dbbook)
    dbbook.current_tx_id = text.id
    db_session.add(dbbook)
    db_session.add(text)
    db_session.commit()

    paragraphs = get_paragraphs(text.text, text.book.language)

    return paragraphs



"""
TokenLocator tests.
"""

from lute.read.render.renderable_calculator import TokenLocator


def test_token_locator_scenario(english):
    """
    Run test scenarios.
    """

    zws = "\u200B"
    cases = [
        # Cases are of the following form:
        # ( tokens,
        #   word sought
        #   matches [ [ as_found_in_text, position ], ... ])
        # Finds b
        (["a", "b", "c", "d"], "b", [{"text": "b", "index": 1}]),
        # Case doesn't matter
        # The original case is returned
        (["A", "B", "C", "D"], "b", [{"text": "B", "index": 1}]),
        # Original case returned
        (["a", "b", "c", "d"], "B", [{"text": "b", "index": 1}]),
        # No match
        (["a", "bb", "c", "d"], "B", []),
        # Found in multiple places
        (
            ["b", "b", "c", "d"],
            "b",
            [{"text": "b", "index": 0}, {"text": "b", "index": 1}],
        ),
        # Multiword, found in multiple
        (
            ["b", "B", "b", "d"],
            f"b{zws}b",
            [{"text": f"b{zws}B", "index": 0}, {"text": f"B{zws}b", "index": 1}],
        ),
        # Multiword, found in multiple
        (
            ["b", "B", "c", "b", "b", "x", "b"],
            f"b{zws}b",
            [{"text": f"b{zws}B", "index": 0}, {"text": f"b{zws}b", "index": 3}],
        ),
        (["a", " ", "cat", " ", "here"], "cat", [{"text": "cat", "index": 2}]),
        (["a", " ", "CAT", " ", "here"], "cat", [{"text": "CAT", "index": 2}]),
        # No match
        (["a", " ", "CAT", " ", "here"], "ca", []),
        # No match
        (["b", "b", "c", "d"], "x", []),
    ]

    casenum = 0
    for tokens, word, expected in cases:
        casenum += 1
        sentence = TokenLocator.make_string(tokens)
        tocloc = TokenLocator(english, sentence)
        actual = tocloc.locate_string(word)
        msg = f"case {casenum} - tokens: {', '.join(tokens)}; word: {word}"
        assert actual == expected, msg



"""
Flask-wtf forms.
"""

import json
from flask import request
from flask_wtf import FlaskForm
from wtforms import (
    StringField,
    SelectField,
    RadioField,
    TextAreaField,
    HiddenField,
    BooleanField,
)
from wtforms import ValidationError
from wtforms.validators import DataRequired

from lute.models.language import Language
from lute.models.term import Term


class TermForm(FlaskForm):
    """
    Term.
    """

    language_id = SelectField("Language", coerce=int)

    original_text = HiddenField("OriginalText")
    text = StringField(
        "Text", validators=[DataRequired()], render_kw={"placeholder": "Term"}
    )
    parentslist = StringField("Parents")

    translation = TextAreaField("Translation", render_kw={"placeholder": "Translation"})
    romanization = StringField(
        "Romanization", render_kw={"placeholder": "Pronunciation"}
    )

    status_choices = [(1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (99, "Wkn"), (98, "Ign")]
    status = RadioField("Status", choices=status_choices)
    sync_status = BooleanField("Link to parent")

    termtagslist = StringField("term_tags")

    current_image = HiddenField("current_image")

    def __init__(self, *args, **kwargs):
        "Call the constructor of the superclass (FlaskForm)"
        super().__init__(*args, **kwargs)
        term = kwargs.get("obj")

        def _data(arr):
            "Get data in proper format for tagify."
            return json.dumps([{"value": p} for p in arr])

        self.parentslist.data = _data(term.parents)
        self.termtagslist.data = _data(term.term_tags)

        if request.method == "POST":
            self.parentslist.data = request.form.get("parentslist", "")
            self.termtagslist.data = request.form.get("termtagslist", "")

    def populate_obj(self, obj):
        "Call the populate_obj method from the parent class, then mine."
        super().populate_obj(obj)

        def _values(field_data):
            "Convert field data to array."
            ret = []
            if field_data:
                ret = [h["value"] for h in json.loads(field_data)]
            return ret

        obj.parents = _values(self.parentslist.data)
        obj.term_tags = _values(self.termtagslist.data)

    def validate_language_id(self, field):  # pylint: disable=unused-argument
        "Language must be set."
        if self.language_id.data in (None, 0):
            raise ValidationError("Please select a language")

    def validate_text(self, field):  # pylint: disable=unused-argument
        "Throw if form text changes from the original or is a dup."
        # Don't continue if the language isn't set.
        if self.language_id.data in (None, 0):
            return
        langid = int(self.language_id.data)
        lang = Language.find(langid)
        if lang is None:
            return

        if self.original_text.data in ("", None):
            # New term.
            spec = Term(lang, self.text.data)
            checkdup = Term.find_by_spec(spec)
            if checkdup is None:
                # Not a dup.
                return
            # Is a dup.
            raise ValidationError("Term already exists")

        if self.text.data == self.original_text.data:
            return
        langid = int(self.language_id.data)
        newterm = Term(lang, self.text.data)
        origterm = Term(lang, self.original_text.data)
        if newterm.text_lc != origterm.text_lc:
            raise ValidationError("Can only change term case")



"""
Step defs for term_parent_mapping.feature.
"""
# pylint: disable=missing-function-docstring

import os
import tempfile
import pytest

from pytest_bdd import given, then, scenarios, parsers

from lute.db import db
from lute.models.term import Term

from lute.term_parent_map.service import import_file, BadImportFileError

from tests.utils import add_terms
from tests.dbasserts import assert_sql_result


# The content of the file for the current test.
content = None

# The current language.
language = None

scenarios("term_parent_mapping.feature")


@given("demo data")
def given_demo_data(app_context):
    "Calling app_context loads the demo data."


@given("language Spanish")
def lang_spanish(spanish):
    "Do-nothing step, just for clarity."
    global language  # pylint: disable=global-statement
    language = spanish


@given(parsers.parse("terms:\n{terms}"))
def given_terms(terms):
    terms = terms.split("\n")
    add_terms(language, terms)


@given(parsers.parse("import file:\n{newcontent}"))
def given_file(newcontent):
    global content  # pylint: disable=global-statement
    content = newcontent


@given("empty import file")
def given_empty_file():
    global content  # pylint: disable=global-statement
    content = ""


@given(parsers.parse('"{parent}" is parent of "{child}"'))
def given_parent(parent, child):
    spec = Term(language, parent)
    p = Term.find_by_spec(spec)
    spec = Term(language, child)
    c = Term.find_by_spec(spec)
    c.parents.append(p)
    db.session.add(p)
    db.session.add(c)
    db.session.commit()


@then(parsers.parse("import should succeed with {created} created, {updated} updated"))
def succeed_with_status(created, updated):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)

    stats = import_file(language, path)
    os.remove(path)

    assert stats["created"] == int(created), "created"
    assert stats["updated"] == int(updated), "updated"


@then(parsers.parse("import should fail with message:\n{message}"))
def fail_with_message(message):
    fd, path = tempfile.mkstemp()
    with os.fdopen(fd, "w") as tmp:
        # do stuff with temp file
        tmp.write(content)
    with pytest.raises(BadImportFileError, match=message):
        import_file(language, path)
    os.remove(path)


@then(parsers.parse("words table should contain:\n{text_lc_content}"))
def then_words_table_contains_WoTextLC(text_lc_content):
    expected = []
    if text_lc_content != "-":
        expected = text_lc_content.split("\n")
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, expected)


@then(parsers.parse("parents should be:\n{expected}"))
def then_parents(expected):
    sql = """
    select p.WoTextLC, c.WoTextLC
    from words p
    inner join wordparents on WpParentWoID=p.WoID
    inner join words c on c.WoID = WpWoID
    order by p.WoTextLC, c.WoTextLC
    """
    if expected == "-":
        expected = []
    else:
        expected = expected.split("\n")
    assert_sql_result(sql, expected)


@then(parsers.parse('sql "{sql}" should return:\n{expected}'))
def then_sql_returns(sql, expected):
    expected = expected.split("\n")
    assert_sql_result(sql, expected)


@then(parsers.parse('"{txt}" flash message should be:\n{msg}'))
def then_flash(txt, msg):
    expected = None
    if msg != "-":
        expected = msg
    spec = Term(language, txt)
    t = Term.find_by_spec(spec)
    assert t.get_flash_message() == expected, "flash"



"""
JapaneseParser tests.
"""

from lute.parse.mecab_parser import JapaneseParser
from lute.models.term import Term
from lute.models.setting import UserSetting
from lute.db import db
from lute.parse.base import ParsedToken


def test_token_count(japanese):
    "token_count checks."
    cases = [("私", 1), ("元気", 1), ("です", 1), ("元気です", 2), ("元気です私", 3)]
    for text, expected_count in cases:
        t = Term(japanese, text)
        assert t.token_count == expected_count, text
        assert t.text_lc == t.text, "case"


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = JapaneseParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def test_end_of_sentence_stored_in_parsed_tokens(japanese):
    "ParsedToken is marked as EOS=True at ends of sentences."
    s = "元気.元気?元気!\n元気。元気？元気！"

    expected = [
        ("元気", True),
        (".", False, True),
        ("元気", True),
        ("?", False, True),
        ("元気", True),
        ("!", False, True),
        ("¶", False, True),
        ("元気", True),
        ("。", False, True),
        ("元気", True),
        ("？", False, True),
        ("元気", True),
        ("！", False, True),
        ("¶", False, True),
    ]
    assert_tokens_equals(s, japanese, expected)


def test_readings(app_context):
    """
    Parser returns readings if they add value.
    """
    p = JapaneseParser()

    # Don't bother giving reading for a few cases
    no_reading = ["NHK", "ツヨイ", "どちら"]  # roman  # only katakana  # only hiragana

    for c in no_reading:
        assert p.get_reading(c) is None, c

    zws = "\u200B"
    cases = [
        ("強い", "ツヨイ"),
        ("二人", "ニニン"),  # ah well, not perfect :-)
        ("強いか", "ツヨイカ"),
        (f"強い{zws}か", f"ツヨイ{zws}カ"),  # zero-width-space ignored
    ]

    for c in cases:
        assert p.get_reading(c[0]) == c[1], c[0]


def test_reading_setting(app_context):
    "Return reading matching user setting."
    cases = {
        "katakana": "ツヨイ",
        "hiragana": "つよい",
        "alphabet": "tsuyoi",
    }
    p = JapaneseParser()
    for k, v in cases.items():
        UserSetting.set_value("japanese_reading", k)
        db.session.commit()
        assert p.get_reading("強い") == v, k



"""
Theme service tests.
"""

import lute.themes.service as svc
from lute.db import db
from lute.models.setting import UserSetting


def test_list_themes():
    "Smoke test only."
    lst = svc.list_themes()
    assert len(lst) > 0, "have themes"
    assert lst[0][0] == "-", "No theme"
    assert lst[0][1] == "(default)"

    assert ("Apple_Books.css", "Apple Books") in lst


def test_default_theme_is_blank_css(app_context):
    "UserSetting starts off with blank css."
    assert UserSetting.get_value("current_theme") == "-"
    assert svc.get_current_css() == "", "Default = empty string."


def test_bad_setting_returns_blank_css(app_context):
    "Just in case."
    UserSetting.set_value("current_theme", "_missing_file.css")
    db.session.commit()
    assert svc.get_current_css() == "", "Missing = empty string."


def test_setting_a_theme_returns_its_css(app_context):
    "User choice is used."
    UserSetting.set_value("current_theme", "Apple_Books.css")
    db.session.commit()
    assert "Georgia" in svc.get_current_css(), "font specified"


def test_next_theme_cycles_themes(app_context):
    """
    Users should be able to move the 'next' theme quickly
    while reading, via a hotkey.
    """
    lst = svc.list_themes()
    assert UserSetting.get_value("current_theme") == lst[0][0]
    svc.next_theme()
    assert UserSetting.get_value("current_theme") == lst[1][0]
    for i in range(0, len(lst) + 10):  # pylint: disable=consider-using-enumerate
        print(i)
        svc.next_theme()
        svc.next_theme()
    # OK



"""
Backup routes.

Backup settings form management, and running backups.
"""

import traceback
from flask import Blueprint, current_app, render_template, request, jsonify, redirect
from lute.models.setting import BackupSettings
from lute.backup.service import create_backup, skip_this_backup


bp = Blueprint("backup", __name__, url_prefix="/backup")


@bp.route("/backup", methods=["GET"])
def backup():
    """
    Endpoint called from front page.

    With extra arg 'type' for manual.
    """
    backuptype = "automatic"
    if "type" in request.args:
        backuptype = "manual"

    settings = BackupSettings.get_backup_settings()
    return render_template(
        "backup/backup.html", backup_folder=settings.backup_dir, backuptype=backuptype
    )


@bp.route("/do_backup", methods=["POST"])
def do_backup():
    """
    Ajax endpoint called from backup.html.
    """
    backuptype = "automatic"
    prms = request.form.to_dict()
    if "type" in prms:
        backuptype = prms["type"]

    c = current_app.env_config
    settings = BackupSettings.get_backup_settings()
    is_manual = backuptype.lower() == "manual"
    try:
        f = create_backup(c, settings, is_manual=is_manual)
        return jsonify(f)
    except Exception as e:  # pylint: disable=broad-exception-caught
        tb = traceback.format_exc()
        return jsonify({"errmsg": str(e) + " -- " + tb}), 500


@bp.route("/skip_this_backup", methods=["GET"])
def handle_skip_this_backup():
    "Update last backup date so backup not attempted again."
    skip_this_backup()
    return redirect("/", 302)



"""
/term routes.
"""

import os
import csv
from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    current_app,
    send_file,
)
from lute.models.language import Language
from lute.models.term import Term as DBTerm
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.term.datatables import get_data_tables_list
from lute.term.model import Repository, Term
from lute.db import db
from lute.term.forms import TermForm
import lute.utils.formutils

bp = Blueprint("term", __name__, url_prefix="/term")


@bp.route("/index", defaults={"search": None}, methods=["GET"])
@bp.route("/index/<search>", methods=["GET"])
def index(search):
    "Index page."
    DBTerm.delete_empty_images()
    languages = db.session.query(Language).order_by(Language.name).all()
    langopts = [(lang.id, lang.name) for lang in languages]
    langopts = [(0, "(all)")] + langopts
    return render_template(
        "term/index.html", initial_search=search, language_options=langopts
    )


def _load_term_custom_filters(request_form, parameters):
    "Manually add filters that the DataTablesFlaskParamParser doesn't know about."
    filter_param_names = [
        "filtLanguage",
        "filtParentsOnly",
        "filtAgeMin",
        "filtAgeMax",
        "filtStatusMin",
        "filtStatusMax",
        "filtIncludeIgnored",
    ]
    request_params = request_form.to_dict(flat=True)
    for p in filter_param_names:
        parameters[p] = request_params.get(p)


@bp.route("/datatables", methods=["POST"])
def datatables_active_source():
    "Datatables data for terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    data = get_data_tables_list(parameters)
    return jsonify(data)


@bp.route("/export_terms", methods=["POST"])
def export_terms():
    "Generate export file of terms."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    parameters["length"] = 1000000
    outfile = os.path.join(current_app.env_config.temppath, "export_terms.csv")
    data = get_data_tables_list(parameters)
    render_data = data["data"]

    # Fields as returned from the datatables query.
    headings = [
        "OMIT_Checkbox",
        "term",
        "parent",
        "translation",
        "language",
        "tags",
        "OMIT_status_text",
        "OMIT_WoID",
        "OMIT_LgID",
        "OMIT_ImageSource",
        "status",
        "link_status",
        "OMIT_status_text",
    ]
    columns_to_exclude = []
    for i, h in enumerate(headings):
        if h.startswith("OMIT_"):
            columns_to_exclude.append(i)

    output_data = [headings] + render_data
    with open(outfile, "w", encoding="utf-8", newline="") as f:
        csv_writer = csv.writer(f)
        for row in output_data:
            filtered_row = [
                value for i, value in enumerate(row) if i not in columns_to_exclude
            ]
            csv_writer.writerow(filtered_row)

    return send_file(outfile, as_attachment=True, download_name="Terms.csv")


def handle_term_form(
    term, repo, form_template_name, return_on_success, embedded_in_reading_frame=False
):
    """
    Handle a form post.

    This is used here and in read.routes -- read.routes.term_form
    lives in an iframe in the reading frames and returns a different
    template on success.
    """
    form = TermForm(obj=term)
    # parents = [{"value": p} for p in term.parents]
    # form.parentslist.data = json.dumps(parents)

    # Flash messages get added on things like term imports.
    # The user opening the form is treated as an acknowledgement.
    term.flash_message = None

    form.language_id.choices = lute.utils.formutils.language_choices()

    if form.validate_on_submit():
        form.populate_obj(term)

        # parents_list_data = request.form.get("parentslist", "")
        # parents_list = json.loads(parents_list_data) if parents_list_data else []
        # parents = [h["value"] for h in parents_list]
        # term.parents = parents

        repo.add(term)
        repo.commit()
        return return_on_success

    hide_pronunciation = False
    # pylint: disable=protected-access
    if term._language is not None:
        hide_pronunciation = not term._language.show_romanization

    return render_template(
        form_template_name,
        form=form,
        term=term,
        language_dicts=Language.all_dictionaries(),
        hide_pronunciation=hide_pronunciation,
        tags=repo.get_term_tags(),
        embedded_in_reading_frame=embedded_in_reading_frame,
    )


def _handle_form(term, repo):
    """
    Handle the form post.  Only show lang. selector
    for new terms.
    """
    return handle_term_form(
        term, repo, "/term/formframes.html", redirect("/term/index", 302)
    )


@bp.route("/edit/<int:termid>", methods=["GET", "POST"])
def edit(termid):
    """
    Edit a term.
    """
    repo = Repository(db)
    term = repo.load(termid)
    return _handle_form(term, repo)


@bp.route("/editbytext/<int:langid>/<text>", methods=["GET", "POST"])
def edit_by_text(langid, text):
    """
    Edit a term.
    """
    repo = Repository(db)
    term = repo.find_or_new(langid, text)
    return _handle_form(term, repo)


@bp.route("/new", methods=["GET", "POST"])
def new():
    """
    Create a term.
    """
    repo = Repository(db)
    term = Term()
    return _handle_form(term, repo)


@bp.route("/search/<text>/<int:langid>", methods=["GET"])
def search_by_text_in_language(text, langid):
    "JSON data for parent data."
    if text.strip() == "" or langid == 0:
        return []
    repo = Repository(db)
    matches = repo.find_matches(langid, text)

    def _make_entry(t):
        return {
            "id": t.id,
            "text": t.text,
            "translation": t.translation,
            "status": t.status,
        }

    result = [_make_entry(t) for t in matches]
    return jsonify(result)


@bp.route("/sentences/<int:langid>/<text>", methods=["GET"])
def sentences(langid, text):
    "Get sentences for terms."
    repo = Repository(db)
    # Use find_or_new(): if the user clicks on a parent tag
    # in the term form, and the parent does not exist yet, then
    # we're creating a new term.
    t = repo.find_or_new(langid, text)
    refs = repo.find_references(t)

    # Transform data for output, to
    # { "term": [refs], "children": [refs], "parent1": [refs], "parent2" ... }
    refdata = [(f'"{text}"', refs["term"]), (f'"{text}" child terms', refs["children"])]
    for p in refs["parents"]:
        refdata.append((f"\"{p['term']}\"", p["refs"]))

    refcount = sum(len(ref[1]) for ref in refdata)
    return render_template(
        "/term/sentences.html",
        text=text,
        no_references=(refcount == 0),
        references=refdata,
    )


@bp.route("/bulk_update_status", methods=["POST"])
def bulk_update_status():
    """
    Update the statuses.

    json:
    {
      langid: x,
      updates: [ { new_status: 1, terms: [ 'a', ] }, ... }, ]
    }
    """
    repo = Repository(db)

    data = request.get_json()
    language_id = int(data.get("langid"))
    updates = data.get("updates")

    for u in updates:
        new_status = int(u.get("new_status"))
        terms = u.get("terms")
        for t in terms:
            term = repo.find_or_new(language_id, t)
            term.status = new_status
            repo.add(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/bulk_set_parent", methods=["POST"])
def bulk_set_parent():
    "Set the parent for terms."
    data = request.get_json()
    termids = data.get("wordids")
    parenttext = data.get("parenttext")
    parent = None
    repo = Repository(db)
    for tid in termids:
        term = repo.load(int(tid))
        if parent is None:
            parent = repo.find(term.language_id, parenttext)
        if term.parents != [parenttext]:
            term.parents = [parenttext]
            term.status = parent.status
            term.sync_status = True
        repo.add(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/bulk_delete", methods=["POST"])
def bulk_delete():
    "Delete terms."
    data = request.get_json()
    termids = data.get("wordids")
    repo = Repository(db)
    for tid in termids:
        term = repo.load(int(tid))
        repo.delete(term)
    repo.commit()
    return jsonify("ok")


@bp.route("/delete/<int:termid>", methods=["POST"])
def delete(termid):
    """
    Delete a term.
    """
    repo = Repository(db)
    term = repo.load(termid)
    repo.delete(term)
    repo.commit()
    return redirect("/term/index", 302)



"""
Common fixtures used by many tests.
"""

import os
import yaml
import pytest

from lute.config.app_config import AppConfig
from lute.db import db
import lute.db.demo
from lute.app_factory import create_app

from lute.models.language import Language


def pytest_sessionstart(session):  # pylint: disable=unused-argument
    """
    Ensure test config defines a test environment.

    Special configuration for test runs required:

    DBNAME must start with test_

    DATAPATH must be specified: this ensures that the tests don't
    accidentally write into the user_data (which could mess with prod
    data/media etc)
    """
    thisdir = os.path.dirname(os.path.realpath(__file__))
    configfile = os.path.join(thisdir, "..", "lute", "config", "config.yml")

    config = None
    with open(configfile, "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)

    failures = []
    if "DATAPATH" not in config:
        failures.append("DATAPATH not in config file")

    ac = AppConfig(configfile)

    if not ac.is_test_db:
        failures.append("DBNAME in config.yml must start with test_")
    if len(failures) > 0:
        msg = f"Bad config.yml: {', '.join(failures)}"
        pytest.exit(msg)


@pytest.fixture(name="testconfig")
def fixture_config():
    "Config using the app config."
    ac = AppConfig(AppConfig.default_config_filename())
    yield ac


@pytest.fixture(name="app")
def fixture_app():
    """
    A clean instance of the demo database.

    I'm not a _huge_ fan of this because if the app
    is open while tests are running, the app seems to hold
    on to references to the old deleted db ...
    that said, it's much faster to do this than to do a
    "wipe and reload database" on every test run.
    """
    config_file = AppConfig.default_config_filename()
    c = AppConfig(config_file)
    if os.path.exists(c.dbfilename):
        os.unlink(c.dbfilename)
    extra_config = {"WTF_CSRF_ENABLED": False, "TESTING": True}
    app = create_app(config_file, extra_config=extra_config)
    yield app


@pytest.fixture(name="app_context")
def fixture_app_context(app):
    """
    Yields the app context so that tests using the db will work.
    """
    with app.app_context() as c:
        yield c


@pytest.fixture(name="empty_db")
def fixture_empty_db(app_context):
    """
    Wipe the db.
    """
    lute.db.management.delete_all_data()


@pytest.fixture(name="client")
def fixture_demo_client(app):
    """
    Client using demo-data-loaded application.
    """
    return app.test_client()


@pytest.fixture(name="demo_yaml_folder")
def fixture_yaml_folder():
    "Path to the demo files."
    return os.path.join(lute.db.demo.demo_data_path(), "languages")


def _get_language(f):
    """
    Return language from the db if it already exists,
    or create it from the file.
    """
    lang = lute.db.demo.get_demo_language(f)
    db_language = db.session.query(Language).filter(Language.name == lang.name).first()
    if db_language is None:
        return lang
    return db_language


@pytest.fixture(name="test_languages")
def fixture_test_languages(app_context, demo_yaml_folder):
    "Dict of available languages for tests."
    # Hardcoded = good enough.
    langs = ["spanish", "english", "japanese", "turkish", "classical_chinese"]
    ret = {}
    for lang in langs:
        f = os.path.join(demo_yaml_folder, f"{lang}.yaml")
        ret[lang] = _get_language(f)
    yield ret


@pytest.fixture(name="spanish")
def fixture_spanish(test_languages):
    return test_languages["spanish"]


@pytest.fixture(name="english")
def fixture_english(test_languages):
    return test_languages["english"]


@pytest.fixture(name="japanese")
def fixture_japanese(test_languages):
    return test_languages["japanese"]


@pytest.fixture(name="turkish")
def fixture_turkish(test_languages):
    return test_languages["turkish"]


@pytest.fixture(name="classical_chinese")
def fixture_cl_chinese(test_languages):
    return test_languages["classical_chinese"]



"""
BackupManager tests.
"""

import os
from lute.db.setup.main import BackupManager


def test_do_backup(tmp_path):
    """
    Running backup should:
    - create a gz backup file with the timestamp added to the name.
    - delete the oldest files in the backup dir matching the basename pattern.
    - leave any other files alone.
    """
    file_to_backup = tmp_path / "sample.txt"
    backup_dir = tmp_path / "backup"

    # Create the backup directory and a sample file
    os.makedirs(backup_dir)
    with open(file_to_backup, "w", encoding="utf-8") as f:
        f.write("Sample file content")

    backup_count = 3
    bm = BackupManager(file_to_backup, backup_dir, backup_count)

    # Perform initial backups.
    bm.do_backup("2000")
    bm.do_backup("2001")
    bm.do_backup("2002")

    # Next one knocks out the first one.
    bm.do_backup("2003")

    # Check if the number of backups is within the specified limit
    backup_files = list(backup_dir.glob("sample.txt.*.gz"))
    assert len(backup_files) == backup_count

    filenames = [os.path.basename(p) for p in backup_files]
    filenames.sort()
    assert filenames == [
        "sample.txt.2001.gz",
        "sample.txt.2002.gz",
        "sample.txt.2003.gz",
    ]



"""
/book routes.
"""

import os

from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    flash,
)
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.book import service
from lute.book.datatables import get_data_tables_list
from lute.book.forms import NewBookForm, EditBookForm
import lute.utils.formutils
from lute.db import db

from lute.models.book import Book as DBBook
from lute.book.model import Book, Repository


bp = Blueprint("book", __name__, url_prefix="/book")


def datatables_source(is_archived):
    "Get datatables json for books."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    data = get_data_tables_list(parameters, is_archived)
    return jsonify(data)


@bp.route("/datatables/active", methods=["POST"])
def datatables_active_source():
    "Datatables data for active books."
    return datatables_source(False)


@bp.route("/archived", methods=["GET"])
def archived():
    "List archived books."
    return render_template("book/index.html", status="Archived")


# Archived must be capitalized, or the ajax call 404's.
@bp.route("/datatables/Archived", methods=["POST"])
def datatables_archived_source():
    "Datatables data for archived books."
    return datatables_source(True)


def _get_file_content(filefielddata):
    """
    Get the content of the file.
    """
    _, ext = os.path.splitext(filefielddata.filename)
    ext = (ext or "").lower()
    if ext == ".txt":
        return service.get_textfile_content(filefielddata)
    if ext == ".epub":
        return service.get_epub_content(filefielddata)
    if ext == ".pdf":
        msg = """
        Note: pdf imports can be inaccurate, due to how PDFs are encoded.
        Please be aware of this while reading.
        """
        flash(msg, "notice")
        return service.get_pdf_content_from_form(filefielddata)
    raise ValueError(f'Unknown file extension "{ext}"')


def _book_from_url(url):
    "Create a new book, or flash an error if can't parse."
    b = Book()
    try:
        b = service.book_from_url(url)
    except service.BookImportException as e:
        flash(e.message, "notice")
        b = Book()
    return b


@bp.route("/new", methods=["GET", "POST"])
def new():
    "Create a new book, either from text or from a file."
    b = Book()
    import_url = request.args.get("importurl", "").strip()
    if import_url != "":
        b = _book_from_url(import_url)

    form = NewBookForm(obj=b)
    form.language_id.choices = lute.utils.formutils.language_choices()
    repo = Repository(db)

    if form.validate_on_submit():
        try:
            form.populate_obj(b)
            if form.textfile.data:
                b.text = _get_file_content(form.textfile.data)
            f = form.audiofile.data
            if f:
                b.audio_filename = service.save_audio_file(f)
            book = repo.add(b)
            repo.commit()
            return redirect(f"/read/{book.id}/page/1", 302)
        except service.BookImportException as e:
            flash(e.message, "notice")

    return render_template(
        "book/create_new.html",
        book=b,
        form=form,
        tags=repo.get_book_tags(),
        show_language_selector=True,
    )


@bp.route("/edit/<int:bookid>", methods=["GET", "POST"])
def edit(bookid):
    "Edit a book - can only change a few fields."
    repo = Repository(db)
    b = repo.load(bookid)
    form = EditBookForm(obj=b)

    if form.validate_on_submit():
        form.populate_obj(b)
        f = form.audiofile.data
        if f:
            b.audio_filename = service.save_audio_file(f)
            b.audio_bookmarks = None
            b.audio_current_pos = None
        repo.add(b)
        repo.commit()
        flash(f"{b.title} updated.")
        return redirect("/", 302)

    return render_template(
        "book/edit.html", book=b, form=form, tags=repo.get_book_tags()
    )


@bp.route("/import_webpage", methods=["GET", "POST"])
def import_webpage():
    return render_template("book/import_webpage.html")


@bp.route("/archive/<int:bookid>", methods=["POST"])
def archive(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    b.archived = True
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/unarchive/<int:bookid>", methods=["POST"])
def unarchive(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    b.archived = False
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/delete/<int:bookid>", methods=["POST"])
def delete(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    db.session.delete(b)
    db.session.commit()
    return redirect("/", 302)



"""
Flask-wtf forms.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField, SelectField
from wtforms.validators import DataRequired


class LanguageForm(FlaskForm):
    """
    Language.
    """

    name = StringField("Name", validators=[DataRequired()])
    dict_1_uri = StringField("Dictionary 1", validators=[DataRequired()])
    dict_2_uri = StringField("Dictionary 2")
    sentence_translate_uri = StringField(
        "Sentence translation", validators=[DataRequired()]
    )
    show_romanization = BooleanField("Show Pronunciation field")
    right_to_left = BooleanField("Right-to-left")

    # Note!  The choices have to be set in the routes!
    # I originally had "choices=lute.parse.registry.supported_parsers()",
    # but it never worked: the Japanese mecab parser was excluded.
    # Possible coder error, not sure, but setting the choices at
    # form creation time works.
    parser_type = SelectField("Parse as", choices=[("tbd", "tbd")])

    character_substitutions = StringField("Character substitutions")

    regexp_split_sentences = StringField(
        "Split sentences at", validators=[DataRequired()]
    )
    exceptions_split_sentences = StringField("Split sentence exceptions")
    word_characters = StringField("Word characters", validators=[DataRequired()])



"""
Database migration manager.

Runs migrations and logs them to a _migrations table so they aren't
run twice.  Runs repeatable migrations always.
"""

import os
from contextlib import contextmanager


@contextmanager
def _change_directory(new_dir):
    "Change directory, and return to the current dir when done."
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(old_dir)


class SqliteMigrator:
    """
    Sqlite migrator class.

    Follows the principles documented in
    https://github.com/jzohrab/DbMigrator/blob/master/docs/managing_database_changes.md
    """

    def __init__(self, location, repeatable):
        self.location = location
        self.repeatable = repeatable

    def has_migrations(self, conn):
        """
        Return True if have non-applied migrations.
        """
        outstanding = self._get_pending(conn)
        return len(outstanding) > 0

    def _get_pending(self, conn):
        """
        Get all non-applied (one-time) migrations.
        """
        allfiles = []
        with _change_directory(self.location):
            allfiles = [
                os.path.join(self.location, s)
                for s in os.listdir()
                if s.endswith(".sql")
            ]
            allfiles.sort()
            outstanding = [f for f in allfiles if self._should_apply(conn, f)]
        return outstanding

    def do_migration(self, conn):
        """
        Run all migrations, then all repeatable migrations.
        """
        self._process_folder(conn)
        self._process_repeatable(conn)

    def _process_folder(self, conn):
        """
        Run all pending migrations.  Write executed script to
        _migrations table.
        """
        outstanding = self._get_pending(conn)
        for f in outstanding:
            try:
                self._process_file(conn, f)
            except Exception as e:
                msg = str(e)
                print(f"\nFile {f} exception:\n{msg}\n")
                raise e
            self._add_migration_to_database(conn, f)

    def _process_repeatable(self, conn):
        """
        Run all repeatable migrations.
        """
        with _change_directory(self.repeatable):
            files = [
                os.path.join(self.repeatable, f)
                for f in os.listdir()
                if f.endswith(".sql")
            ]
        for f in files:
            try:
                self._process_file(conn, f)
            except Exception as e:
                msg = str(e)
                print(f"\nFile {f} exception:\n{msg}\n")
                raise e

    def _should_apply(self, conn, filename):
        """
        True if a migration hasn't been run yet.

        The file basename (no directory) is stored in the migration table.
        """
        f = os.path.basename(filename)
        sql = f"select count(filename) from _migrations where filename = '{f}'"
        res = conn.execute(sql).fetchone()
        return res[0] == 0

    def _add_migration_to_database(self, conn, filename):
        """
        Track the executed migration in _migrations.
        """
        f = os.path.basename(filename)
        conn.execute("begin transaction")
        conn.execute(f"INSERT INTO _migrations values ('{f}')")
        conn.execute("commit transaction")

    def _process_file(self, conn, f):
        """
        Run the given file.
        """
        with open(f, "r", encoding="utf8") as sql_file:
            commands = sql_file.read()
            self._exec_commands(conn, commands)

    def _exec_commands(self, conn, sql):
        """
        Execute all commands in the given file.
        """
        conn.executescript(sql)



"""
Smoke tests.
"""


def test_smoke_test(chromebrowser, luteclient):
    "Hit the main page, create a book, update a term."
    luteclient.visit("/")
    luteclient.clear_book_filter()
    assert chromebrowser.is_text_present("Lute"), "have main page."
    luteclient.make_book("Hola", "Hola. Adios amigo.", "Spanish")
    assert chromebrowser.title == 'Reading "Hola"', "title"
    assert chromebrowser.is_text_present("Hola")
    assert "Hola/. /Adios/ /amigo/." == luteclient.displayed_text()

    updates = {"translation": "hello", "parents": ["adios", "amigo"]}
    luteclient.click_word_fill_form("Hola", updates)
    luteclient.click_word_fill_form("Adios", {"status": "2", "translation": "goodbye"})

    displayed = luteclient.displayed_text()
    assert "Hola (1)/. /Adios (2)/ /amigo (1)/." == displayed


def test_unsupported_language_not_shown(luteclient, _restore_jp_parser):
    "Missing mecab means no Japanese."
    luteclient.load_demo_stories()

    luteclient.change_parser_registry_key("japanese", "disabled_japanese")
    luteclient.visit("/")
    luteclient.clear_book_filter()
    assert not luteclient.browser.is_text_present("Japanese"), "no Japanese demo book."
    assert luteclient.browser.is_text_present(
        "Tutorial"
    ), "Tutorial is available though."



"""
Language entity.
"""

import re
from sqlalchemy import text, func
from lute.db import db
from lute.parse.registry import get_parser, is_supported


class Language(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Language entity.
    """

    __tablename__ = "languages"

    id = db.Column("LgID", db.SmallInteger, primary_key=True)
    name = db.Column("LgName", db.String(40))
    dict_1_uri = db.Column("LgDict1URI", db.String(200))
    dict_2_uri = db.Column("LgDict2URI", db.String(200))
    sentence_translate_uri = db.Column("LgGoogleTranslateURI", db.String(200))
    character_substitutions = db.Column("LgCharacterSubstitutions", db.String(500))
    regexp_split_sentences = db.Column("LgRegexpSplitSentences", db.String(500))
    exceptions_split_sentences = db.Column("LgExceptionsSplitSentences", db.String(500))
    _word_characters = db.Column("LgRegexpWordCharacters", db.String(500))
    remove_spaces = db.Column("LgRemoveSpaces", db.Boolean)
    split_each_char = db.Column("LgSplitEachChar", db.Boolean)
    right_to_left = db.Column("LgRightToLeft", db.Boolean)
    show_romanization = db.Column("LgShowRomanization", db.Boolean)
    parser_type = db.Column("LgParserType", db.String(20))

    def __init__(self):
        self.character_substitutions = "´='|`='|’='|‘='|...=…|..=‥"
        self.regexp_split_sentences = ".!?"
        self.exceptions_split_sentences = "Mr.|Mrs.|Dr.|[A-Z].|Vd.|Vds."
        self.word_characters = "a-zA-ZÀ-ÖØ-öø-ȳáéíóúÁÉÍÓÚñÑ"
        self.remove_spaces = False
        self.split_each_char = False
        self.right_to_left = False
        self.show_romanization = False
        self.parser_type = "spacedel"

    def __repr__(self):
        return f"<Language {self.id} '{self.name}'>"

    def _get_python_regex_pattern(self, s):
        """
        Old Lute v2 ran in php, so the language word chars regex
        could look like this:

        x{0600}-x{06FF}x{FE70}-x{FEFC}  (where x = backslash-x)

        This needs to be converted to the python equivalent, e.g.

        u0600-u06FFuFE70-uFEFC  (where u = backslash-u)
        """

        def convert_match(match):
            # Convert backslash-x{XXXX} to backslash-uXXXX
            hex_value = match.group(1)
            return f"\\u{hex_value}"

        ret = re.sub(r"\\x{([0-9A-Fa-f]+)}", convert_match, s)
        return ret

    @property
    def word_characters(self):
        return self._get_python_regex_pattern(self._word_characters)

    @word_characters.setter
    def word_characters(self, s):
        self._word_characters = self._get_python_regex_pattern(s)

    @classmethod
    def all_dictionaries(cls):
        """
        All dictionaries for all languages.
        """
        languages = Language.query.all()
        language_data = {}
        for language in languages:
            term_dicts = [language.dict_1_uri, language.dict_2_uri]
            term_dicts = [uri for uri in term_dicts if uri is not None]

            data = {"term": term_dicts, "sentence": language.sentence_translate_uri}

            language_data[language.id] = data
        return language_data

    @staticmethod
    def delete(language):
        """
        Hacky method to delete language and all terms and books
        associated with it.

        There is _certainly_ a better way to do this using
        Sqlalchemy relationships and cascade deletes, but I
        was running into problems with it (things not cascading,
        or warnings ("SAWarning: Object of type <Term> not in
        session, add operation along 'Language.terms' will not
        proceed") during test runs.  It would be nice to have
        a "correct" mapping, but this is good enough for now.

        TODO zzfuture fix: fix Language-Book and -Term mappings.
        """
        sqls = [
            "pragma foreign_keys = ON",
            f"delete from languages where LgID = {language.id}",
        ]
        for s in sqls:
            db.session.execute(text(s))
        db.session.commit()

    @property
    def parser(self):
        return get_parser(self.parser_type)

    @property
    def is_supported(self):
        "True if the language's parser is supported."
        return is_supported(self.parser_type)

    def get_parsed_tokens(self, s):
        return self.parser.get_parsed_tokens(s, self)

    def get_lowercase(self, s) -> str:
        return self.parser.get_lowercase(s)

    @staticmethod
    def find(language_id):
        "Get by ID."
        return db.session.query(Language).filter(Language.id == language_id).first()

    @staticmethod
    def find_by_name(name):
        "Get by name."
        return (
            db.session.query(Language)
            .filter(func.lower(Language.name) == func.lower(name))
            .first()
        )



"""
Parsing for space-delimited languages.

The parser uses some Language settings (e.g., word characters) to
perform the actual parsing.

Includes classes:

- SpaceDelimitedParser
- Turkish
"""

import re
from typing import List
from lute.parse.base import ParsedToken, AbstractParser


class SpaceDelimitedParser(AbstractParser):
    """
    A general parser for space-delimited languages,
    such as English, French, Spanish ... etc.
    """

    @classmethod
    def name(cls):
        return "Space Delimited"

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        "Return parsed tokens."
        clean_text = re.sub(r" +", " ", text)
        zws = chr(0x200B)  # zero-width space
        clean_text = clean_text.replace(zws, "")
        return self._parse_to_tokens(clean_text, language)

    def preg_match_capture(self, pattern, subject):
        """
        Return the matched text and their start positions in the subject.

        E.g. search for r'cat' in "there is a CAT and a Cat" returns:
        [['CAT', 11], ['Cat', 21]]
        """
        matches = re.finditer(pattern, subject, flags=re.IGNORECASE)
        result = [[match.group(), match.start()] for match in matches]
        return result

    def _parse_to_tokens(self, text: str, lang):
        """
        Returns ParsedToken array for given language.
        """
        replacements = lang.character_substitutions.split("|")
        for replacement in replacements:
            fromto = replacement.strip().split("=")
            if len(fromto) >= 2:
                rfrom = fromto[0].strip()
                rto = fromto[1].strip()
                text = text.replace(rfrom, rto)

        text = text.replace("\r\n", "\n")
        text = text.replace("{", "[")
        text = text.replace("}", "]")

        tokens = []
        paras = text.split("\n")
        pcount = len(paras)
        for i, para in enumerate(paras):
            self.parse_para(para, lang, tokens)
            if i != (pcount - 1):
                tokens.append(ParsedToken("¶", False, True))

        return tokens

    def parse_para(self, text: str, lang, tokens: List[ParsedToken]):
        """
        Parse a string, appending the tokens to the list of tokens.
        """
        termchar = lang.word_characters
        if termchar.strip() == "":
            raise RuntimeError(
                f"Language {lang.name} has invalid Word Characters specification."
            )

        splitex = lang.exceptions_split_sentences.replace(".", "\\.")
        pattern = rf"({splitex}|[{termchar}]*)"
        if splitex.strip() == "":
            pattern = rf"([{termchar}]*)"

        m = self.preg_match_capture(pattern, text)
        wordtoks = list(filter(lambda t: t[0] != "", m))

        def add_non_words(s):
            """
            Add non-word token s to the list of tokens.  If s
            matches any of the split_sentence values, mark it as an
            end-of-sentence.
            """
            if not s:
                return
            pattern = f"[{re.escape(lang.regexp_split_sentences)}]"
            has_eos = False
            if pattern != "[]":  # Should never happen, but ...
                allmatches = self.preg_match_capture(pattern, s)
                has_eos = len(allmatches) > 0
            tokens.append(ParsedToken(s, False, has_eos))

        # For each wordtok, add all non-words before the wordtok, and
        # then add the wordtok.
        pos = 0
        for wt in wordtoks:
            w = wt[0]
            wp = wt[1]
            s = text[pos:wp]
            add_non_words(s)
            tokens.append(ParsedToken(w, True, False))
            pos = wp + len(w)

        # Add anything left over.
        s = text[pos:]
        add_non_words(s)


class TurkishParser(SpaceDelimitedParser):
    "Parser to handle Turkish parsing fun."

    @classmethod
    def name(cls):
        return "Turkish"

    def get_lowercase(self, text):
        "Handle the funny turkish i variants."
        for caps, lower in {"İ": "i", "I": "ı"}.items():
            text = text.replace(caps, lower)
        return text.lower()



"""
User audio routes.

User audio files are stored in the database in books table.
"""

import os
from flask import Blueprint, send_file, current_app
from lute.models.book import Book

bp = Blueprint("useraudio", __name__, url_prefix="/useraudio")


@bp.route("/stream/<int:bookid>", methods=["GET"])
def stream(bookid):
    "Serve the audio, no caching."
    dirname = current_app.env_config.useraudiopath
    b = Book.find(bookid)
    fname = os.path.join(dirname, b.audio_filename)
    return send_file(fname, as_attachment=True, max_age=0)



"""
Term entity.
"""

from sqlalchemy import text as sqltext, and_
from lute.db import db

wordparents = db.Table(
    "wordparents",
    db.Model.metadata,
    db.Column("WpWoID", db.Integer, db.ForeignKey("words.WoID")),
    db.Column("WpParentWoID", db.Integer, db.ForeignKey("words.WoID")),
)


class TermImage(db.Model):
    "Term images."
    __tablename__ = "wordimages"

    id = db.Column("WiID", db.Integer, primary_key=True)
    term_id = db.Column(
        "WiWoID", db.Integer, db.ForeignKey("words.WoID"), nullable=False
    )
    term = db.relationship("Term", back_populates="images")
    source = db.Column("WiSource", db.String(500))


class TermFlashMessage(db.Model):
    "Term flash messages."
    __tablename__ = "wordflashmessages"

    id = db.Column("WfID", db.Integer, primary_key=True)
    message = db.Column("WfMessage", db.String(200))
    term_id = db.Column(
        db.Integer, db.ForeignKey("words.WoID"), name="WfWoID", nullable=False
    )
    term = db.relationship("Term", back_populates="term_flash_message", uselist=False)


wordtags = db.Table(
    "wordtags",
    db.Model.metadata,
    db.Column("WtTgID", db.Integer, db.ForeignKey("tags.TgID")),
    db.Column("WtWoID", db.Integer, db.ForeignKey("words.WoID")),
)


class TermTag(db.Model):
    "Term tags."
    __tablename__ = "tags"

    id = db.Column("TgID", db.Integer, primary_key=True)
    text = db.Column("TgText", db.String(20))
    _comment = db.Column("TgComment", db.String(200))

    def __init__(self, text, comment=None):
        self.text = text
        self.comment = comment

    @property
    def comment(self):
        "Comment getter."
        return self._comment

    # TODO zzfuture fix: TgComment should be nullable
    # The current schema has the TgComment NOT NULL default '',
    # which was a legacy copy over from LWT. Really, this should
    # be a nullable column, and all blanks should be NULL.
    # Minor change, a nice-to-have.
    @comment.setter
    def comment(self, c):
        "Set cleaned comment."
        self._comment = c if c is not None else ""

    @staticmethod
    def find(termtag_id):
        "Get by ID."
        return db.session.query(TermTag).filter(TermTag.id == termtag_id).first()

    @staticmethod
    def find_by_text(text):
        "Find a tag by text, or None if not found."
        return db.session.query(TermTag).filter(TermTag.text == text).first()

    @staticmethod
    def find_or_create_by_text(text):
        "Return tag or create one."
        ret = TermTag.find_by_text(text)
        if ret is not None:
            return ret
        return TermTag(text)


class TermTextChangedException(Exception):
    """
    Terms cannot change their text once saved,
    except for the text case.
    """


class Term(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Term entity.
    """

    __tablename__ = "words"

    id = db.Column("WoID", db.SmallInteger, primary_key=True)
    language_id = db.Column("WoLgID", db.Integer, db.ForeignKey("languages.LgID"))

    # Text should only be set through setters.
    _text = db.Column("WoText", db.String(250))

    # text_lc shouldn't be touched (it's changed when term.text is
    # set), but it's public here to allow for its access during
    # queries (eg in lute.read.service.  This can probably be
    # improved, not sure how at the moment.
    text_lc = db.Column("WoTextLC", db.String(250))

    status = db.Column("WoStatus", db.Integer)
    translation = db.Column("WoTranslation", db.String(500))
    romanization = db.Column("WoRomanization", db.String(100))
    token_count = db.Column("WoTokenCount", db.Integer)
    sync_status = db.Column("WoSyncStatus", db.Boolean)

    language = db.relationship("Language")
    term_tags = db.relationship("TermTag", secondary="wordtags")
    parents = db.relationship(
        "Term",
        secondary="wordparents",
        primaryjoin="Term.id == wordparents.c.WpWoID",
        secondaryjoin="Term.id == wordparents.c.WpParentWoID",
        back_populates="children",
    )
    children = db.relationship(
        "Term",
        secondary="wordparents",
        primaryjoin="Term.id == wordparents.c.WpParentWoID",
        secondaryjoin="Term.id == wordparents.c.WpWoID",
        back_populates="parents",
    )
    images = db.relationship(
        "TermImage",
        back_populates="term",
        lazy="subquery",
        cascade="all, delete-orphan",
    )
    term_flash_message = db.relationship(
        "TermFlashMessage",
        uselist=False,
        back_populates="term",
        cascade="all, delete-orphan",
    )

    def __init__(self, language=None, text=None):
        self.status = 1
        self.translation = None
        self.romanization = None
        self.sync_status = False
        self.term_tags = []
        self.parents = []
        self.children = []
        self.images = []
        if language is not None:
            self.language = language
        if text is not None:
            self.text = text

    def __repr__(self):
        return f"<Term {self.id} '{self.text}'>"

    def __eq__(self, other):
        def eqrep(term):
            return f"{term.language.name} {term.text}"

        if isinstance(other, Term):
            return eqrep(self) == eqrep(other)
        return False

    @property
    def text(self):
        "Get the text."
        return self._text

    @text.setter
    def text(self, textstring):
        "Set the text, textlc, and token count."
        if self.language is None:
            raise RuntimeError("Must set term language before setting text")

        # Clean up encoding cruft.
        t = textstring.strip()
        zws = "\u200B"  # zero-width space
        t = t.replace(zws, "")
        nbsp = "\u00A0"  # non-breaking space
        t = t.replace(nbsp, " ")

        lang = self.language
        tokens = lang.get_parsed_tokens(t)

        # Terms can't contain paragraph markers.
        tokens = [tok for tok in tokens if tok.token != "¶"]
        tok_strings = [tok.token for tok in tokens]

        t = zws.join(tok_strings)
        old_text_lc = self.text_lc
        new_text_lc = lang.get_lowercase(t)

        text_changed = old_text_lc is not None and new_text_lc != old_text_lc
        if self.id is not None and text_changed:
            msg = (
                f"Cannot change text of term '{self.text}' (id = {self.id}) once saved."
            )
            raise TermTextChangedException(msg)

        self._text = t
        self.text_lc = new_text_lc
        self._calc_token_count()

    def _calc_token_count(self):
        "Tokens are separated by zero-width space."
        token_count = 0
        if self._text is not None:
            zws = "\u200B"  # zero-width space
            parts = self._text.split(zws)
            token_count = len(parts)
        self.token_count = token_count

    def remove_all_term_tags(self):
        self.term_tags = []

    def add_term_tag(self, term_tag):
        if term_tag not in self.term_tags:
            self.term_tags.append(term_tag)

    def remove_term_tag(self, term_tag):
        self.term_tags.remove(term_tag)

    def remove_all_parents(self):
        self.parents = []

    def add_parent(self, parent):
        """
        Add valid parent, term cannot be its own parent.
        """
        if self == parent:
            return
        if parent not in self.parents:
            self.parents.append(parent)

    def get_current_image(self, strip_jpeg=True):
        "Get the current (first) image for the term."
        if len(self.images) == 0:
            return None
        i = self.images[0]

        src = i.source

        if not strip_jpeg:
            return src

        # Ugly hack: we have to remove the .jpeg at the end because
        # Flask doesn't handle params with periods.
        return src.replace(".jpeg", "")

    def set_current_image(self, s):
        "Set the current image for this term."
        while len(self.images) > 0:
            self.images.pop(0)
        if (s or "").strip() != "":
            ti = TermImage()
            ti.term = self
            ti.source = s.strip()
            self.images.append(ti)

    @staticmethod
    def delete_empty_images():
        """
        Data clean-up: delete empty images.

        The code was leaving empty images in the db, which are obviously no good.
        This is a hack to clean up the data.
        """
        sql = "delete from wordimages where trim(WiSource) = ''"
        db.session.execute(sqltext(sql))
        db.session.commit()

    def get_flash_message(self):
        "Get the flash message."
        if not self.term_flash_message:
            return None
        return self.term_flash_message.message

    def set_flash_message(self, m):
        "Set a flash message to be shown at some point in the future."
        tfm = self.term_flash_message
        if not tfm:
            tfm = TermFlashMessage()
            self.term_flash_message = tfm
        tfm.message = m

    def pop_flash_message(self):
        "Get the flash message, and remove it from this term."
        if not self.term_flash_message:
            return None
        m = self.term_flash_message.message
        self.term_flash_message = None
        return m

    @staticmethod
    def find(term_id):
        "Get by ID."
        return db.session.query(Term).filter(Term.id == term_id).first()

    @staticmethod
    def find_by_spec(spec):
        """
        Find by the given spec term's language ID and text.
        Returns None if not found.
        """
        langid = spec.language.id
        text_lc = spec.text_lc
        query = db.session.query(Term).filter(
            and_(Term.language_id == langid, Term.text_lc == text_lc)
        )
        terms = query.all()
        if not terms:
            return None
        return terms[0]


class Status:
    """
    Term statuses.
    """

    UNKNOWN = 0
    WELLKNOWN = 99
    IGNORED = 98



"""
Mapping parents.
"""

import os
from flask import Blueprint, current_app, render_template, flash, redirect, send_file
from wtforms import SelectField, ValidationError
from wtforms.validators import DataRequired
from flask_wtf import FlaskForm
from flask_wtf.file import FileField
from sqlalchemy import asc
from lute.db import db
from lute.models.book import Book
from lute.models.language import Language
import lute.utils.formutils
from lute.term_parent_map.service import (
    import_file,
    BadImportFileError,
    export_unknown_terms,
    export_terms_without_parents,
)


bp = Blueprint("term_parent_map", __name__, url_prefix="/term_parent_map")


class TermParentMapImportForm(FlaskForm):
    "Form for imports."
    language_id = SelectField("Language", coerce=int)
    text_file = FileField("Text File", validators=[DataRequired()])

    def validate_language_id(self, field):  # pylint: disable=unused-argument
        "Language must be set."
        if self.language_id.data in (None, 0):
            raise ValidationError("Please select a language")


@bp.route("/index", methods=["GET", "POST"])
def index():
    """
    Show books and languages, process import post.
    """
    form = TermParentMapImportForm()
    form.language_id.choices = lute.utils.formutils.language_choices()

    if form.validate_on_submit():
        text_file = form.text_file.data
        language = db.session.get(Language, form.language_id.data)
        if text_file:
            temp_file_name = os.path.join(
                current_app.env_config.temppath, "import_parent_map.txt"
            )
            text_file.save(temp_file_name)
            try:
                stats = import_file(language, temp_file_name)
                msg = (
                    f"Imported {language.name} mappings: "
                    + f"created {stats['created']} terms, updated {stats['updated']}."
                )
                flash(msg, "notice")
                return redirect("/term_parent_map/index", 302)
            except BadImportFileError as e:
                flash(f"Error on import: {str(e)}", "notice")

    # sqlalchemy _requires_ "== False" for the comparison!
    # pylint: disable=singleton-comparison
    books = (
        db.session.query(Book)
        .filter(Book.archived == False)
        .order_by(asc(Book.title))
        .all()
    )
    languages = db.session.query(Language).order_by(asc(Language.name)).all()
    return render_template(
        "/term_parent_map/index.html", books=books, languages=languages, form=form
    )


@bp.route("/export_book/<int:bookid>", methods=["GET"])
def export_book(bookid):
    "Generate a file and return it."
    outfile = os.path.join(current_app.env_config.temppath, "export_book.txt")
    book = db.session.get(Book, bookid)
    export_unknown_terms(book, outfile)
    return send_file(outfile, as_attachment=True, download_name="unknown_terms.txt")


@bp.route("/export_language/<int:languageid>", methods=["GET"])
def export_language(languageid):
    "Generate a file and return it."
    outfile = os.path.join(current_app.env_config.temppath, "export_lang.txt")
    lang = db.session.get(Language, languageid)
    export_terms_without_parents(lang, outfile)
    return send_file(
        outfile,
        as_attachment=True,
        download_name="terms_without_parents.txt",
    )



"""
Term import.
"""

import os
from flask import Blueprint, current_app, render_template, flash, redirect
from wtforms import BooleanField
from wtforms.validators import DataRequired
from flask_wtf import FlaskForm
from flask_wtf.file import FileField
from lute.termimport.service import import_file, BadImportFileError


bp = Blueprint("termimport", __name__, url_prefix="/termimport")


class TermImportForm(FlaskForm):
    "Form for imports."
    text_file = FileField("Text File", validators=[DataRequired()])
    create_terms = BooleanField("Create new Terms")
    update_terms = BooleanField("Update existing Terms")


@bp.route("/index", methods=["GET", "POST"])
def term_import_index():
    "Import posted file."
    form = TermImportForm()

    if form.validate_on_submit():
        text_file = form.text_file.data
        if text_file:
            temp_file_name = os.path.join(
                current_app.env_config.temppath, "import_terms.txt"
            )
            text_file.save(temp_file_name)
            try:
                stats = import_file(
                    temp_file_name, form.create_terms.data, form.update_terms.data
                )
                c = stats["created"]
                u = stats["updated"]
                s = stats["skipped"]
                flash(
                    f"Imported {c} terms, updated {u} (skipped {s})",
                    "notice",
                )
                return redirect("/term/index", 302)
            except BadImportFileError as e:
                flash(f"Error on import: {str(e)}", "notice")

    return render_template("termimport/index.html", form=form)



"""
TurkishParser tests.
"""

from lute.models.term import Term


def test_downcase(turkish):
    "Turkish has problematic 'i' variants."
    cases = [
        ("CAT", "cat"),  # dummy case
        ("İÇİN", "için"),
        ("IŞIK", "ışık"),
        ("İçin", "için"),
        ("Işık", "ışık"),
    ]

    for text, expected_lcase in cases:
        t = Term(turkish, text)
        assert t.text_lc == expected_lcase, text



"""
Book stats tests.
"""

import pytest

from lute.db import db
from lute.term.model import Term, Repository
from lute.book.stats import get_status_distribution, refresh_stats, mark_stale

from tests.utils import make_text, make_book
from tests.dbasserts import assert_record_count_equals, assert_sql_result


def add_term(lang, s, status):
    "Create and add term."
    term = Term()
    term.language = lang
    term.language_id = lang.id
    term.text = s
    term.status = status
    repo = Repository(db)
    repo.add(term)
    repo.commit()


def scenario(language, fulltext, terms_and_statuses, expected):
    "Run a scenario."
    t = make_text("Hola", fulltext, language)
    b = t.book
    db.session.add(t)
    db.session.add(b)
    db.session.commit()

    for ts in terms_and_statuses:
        add_term(language, ts[0], ts[1])

    stats = get_status_distribution(b)

    assert stats == expected


def test_two_words(spanish):
    scenario(
        spanish,
        "Tengo un gato.  Tengo un perro.",
        [["gato", 1], ["perro", 2]],
        {0: 2, 1: 1, 2: 1, 3: 0, 4: 0, 5: 0, 98: 0, 99: 0},
    )


def test_single_word(spanish):
    scenario(
        spanish,
        "Tengo un gato.  Tengo un perro.",
        [["gato", 3]],
        {0: 3, 1: 0, 2: 0, 3: 1, 4: 0, 5: 0, 98: 0, 99: 0},
    )


def test_with_multiword(spanish):
    scenario(
        spanish,
        "Tengo un gato.  Tengo un perro.",
        [["tengo un", 3]],
        {0: 2, 1: 0, 2: 0, 3: 1, 4: 0, 5: 0, 98: 0, 99: 0},
    )


def test_chinese_no_term_stats(classical_chinese):
    scenario(
        classical_chinese,
        "這是東西",
        [],
        {0: 4, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 98: 0, 99: 0},
    )


def test_chinese_with_terms(classical_chinese):
    scenario(
        classical_chinese,
        "這是東西",
        [["東西", 1]],
        {0: 2, 1: 1, 2: 0, 3: 0, 4: 0, 5: 0, 98: 0, 99: 0},
    )


@pytest.fixture(name="_test_book")
def fixture_make_book(empty_db, spanish):
    "Single page book."
    b = make_book("Hola.", "Hola tengo un gato.", spanish)
    db.session.add(b)
    db.session.commit()
    return b


def add_terms(lang, terms):
    "Create and add term."
    repo = Repository(db)
    for s in terms:
        term = Term()
        term.language = lang
        term.language_id = lang.id
        term.text = s
        repo.add(term)
    repo.commit()


def assert_stats(expected, msg=""):
    "helper."
    sql = """select distinctterms, distinctunknowns,
      unknownpercent, replace(status_distribution, '"', "'") from bookstats"""
    assert_sql_result(sql, expected, msg)


def test_cache_loads_when_prompted(_test_book):
    "Have to call refresh_stats() to load stats."
    assert_record_count_equals("bookstats", 0, "nothing loaded")
    refresh_stats()
    assert_record_count_equals("bookstats", 1, "loaded")


def test_stats_smoke_test(_test_book, spanish):
    "Terms are rendered to count stats."
    add_terms(spanish, ["gato", "TENGO"])
    refresh_stats()
    assert_stats(
        ["4; 2; 50; {'0': 2, '1': 2, '2': 0, '3': 0, '4': 0, '5': 0, '98': 0, '99': 0}"]
    )


def test_stats_calculates_rendered_text(_test_book, spanish):
    "Multiword term counted as one term."
    add_terms(spanish, ["tengo un"])
    refresh_stats()
    assert_stats(
        ["3; 2; 67; {'0': 2, '1': 1, '2': 0, '3': 0, '4': 0, '5': 0, '98': 0, '99': 0}"]
    )


def test_stats_only_update_books_marked_stale(_test_book, spanish):
    "Have to mark book as stale, too expensive otherwise."
    add_terms(spanish, ["gato", "TENGO"])
    refresh_stats()
    assert_stats(
        ["4; 2; 50; {'0': 2, '1': 2, '2': 0, '3': 0, '4': 0, '5': 0, '98': 0, '99': 0}"]
    )

    add_terms(spanish, ["hola"])
    refresh_stats()
    assert_stats(
        [
            "4; 2; 50; {'0': 2, '1': 2, '2': 0, '3': 0, '4': 0, '5': 0, '98': 0, '99': 0}"
        ],
        "not updated",
    )

    mark_stale(_test_book)
    refresh_stats()
    assert_stats(
        [
            "4; 1; 25; {'0': 1, '1': 3, '2': 0, '3': 0, '4': 0, '5': 0, '98': 0, '99': 0}"
        ],
        "updated",
    )



"""
DB Backup tests.
"""

import os
from datetime import datetime
import pytest

from lute.backup.service import (
    create_backup,
    BackupException,
    should_run_auto_backup,
    backup_warning,
)
from lute.models.setting import BackupSettings

# pylint: disable=missing-function-docstring
# Test method names are pretty descriptive already.


@pytest.fixture(name="bkp_dir")
def fixture_backup_directory(testconfig):
    "Create clean backup dir."
    base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    backup_dir = os.path.join(base_dir, "zz_bkp")
    if not os.path.exists(backup_dir):
        os.mkdir(backup_dir)

    # Cleanup prior to test.
    cleanup_directory(backup_dir)
    cleanup_directory(testconfig.userimagespath)

    yield backup_dir

    # Don't clean up after the test, in case want to check manually.
    # cleanup_directories(backup_dir, image_dir)


def cleanup_directory(directory):
    "Delete all files and subfolders."
    for root, dirs, files in os.walk(directory, topdown=False):
        for f in files:
            os.remove(os.path.join(root, f))
        for d in dirs:
            os.rmdir(os.path.join(root, d))


@pytest.fixture(name="backup_settings")
def fixture_backup_settings(app_context, bkp_dir):
    # app_context is passed so that the db session is available.
    ret = BackupSettings.get_backup_settings()
    ret.backup_dir = bkp_dir
    ret.backup_enabled = True
    yield ret


def test_backup_writes_file_to_output_dir(testconfig, bkp_dir, backup_settings):
    file_content = b"imagefile"
    img1 = os.path.join(testconfig.userimagespath, "1")
    if not os.path.exists(img1):
        os.mkdir(img1)
    with open(os.path.join(img1, "file.txt"), "wb") as f:
        f.write(file_content)

    create_backup(testconfig, backup_settings)
    assert len(os.listdir(bkp_dir)) == 2  # db and directory
    assert len(os.listdir(os.path.join(bkp_dir, "userimages_backup", "1"))) == 1
    assert os.path.exists(os.path.join(bkp_dir, "userimages_backup", "1", "file.txt"))


def test_timestamp_added_to_db_name(testconfig, bkp_dir, backup_settings):
    assert os.listdir(bkp_dir) == [], "empty dirs at start"
    create_backup(testconfig, backup_settings)
    dbfile = [f for f in os.listdir(bkp_dir) if f.startswith("lute_backup_2")]
    assert len(dbfile) == 1, "db found"


def test_backup_fails_if_missing_output_dir(testconfig, backup_settings):
    backup_settings.backup_dir = "some_missing_dir"
    with pytest.raises(BackupException, match="Missing directory some_missing_dir"):
        create_backup(testconfig, backup_settings)


def test_user_can_configure_rolling_backup_count(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    for i in range(1, 10):
        create_backup(testconfig, backup_settings, suffix=f"0{i}")
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(8, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_all_manual_backups_are_kept(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    for i in range(1, 10):
        create_backup(testconfig, backup_settings, suffix=f"0{i}", is_manual=True)
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(1, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_last_import_setting_is_updated_on_successful_backup(
    testconfig, backup_settings
):
    assert backup_settings.last_backup_datetime is None, "no backup"
    create_backup(testconfig, backup_settings)
    updated = BackupSettings.get_backup_settings()
    assert updated.last_backup_datetime is not None, "set"


def test_should_not_run_autobackup_if_auto_is_no_or_false(backup_settings):
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = False
    assert should_run_auto_backup(backup_settings) is False


def test_autobackup_returns_true_if_never_backed_up(backup_settings):
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = True
    backup_settings.last_backup_datetime = None
    assert should_run_auto_backup(backup_settings) is True


def test_autobackup_returns_true_if_last_backed_up_over_one_day_ago(backup_settings):
    backup_settings.backup_auto = True
    # backup_settings.backup_warn = True
    curr_datetime = datetime.now()
    one_day_ago = curr_datetime.timestamp() - 24 * 60 * 60

    backup_settings.last_backup_datetime = one_day_ago - 10
    assert should_run_auto_backup(backup_settings) is True

    backup_settings.last_backup_datetime = one_day_ago + 10
    assert should_run_auto_backup(backup_settings) is False


def test_warn_if_last_backup_never_happened_or_is_old(backup_settings):
    curr_datetime = datetime.now()
    one_week_ago = curr_datetime.timestamp() - 7 * 24 * 60 * 60

    backup_settings.backup_warn = True
    backup_settings.last_backup_datetime = None
    assert backup_warning(backup_settings) == "Never backed up."

    backup_settings.last_backup_datetime = one_week_ago + 10
    assert backup_warning(backup_settings) == ""

    backup_settings.last_backup_datetime = one_week_ago - 10
    assert backup_warning(backup_settings) == "Last backup was more than 1 week ago."

    backup_settings.backup_warn = False
    assert backup_warning(backup_settings) == ""



"""
Parsing using MeCab.

Uses natto-py (https://github.com/buruzaemon/natto-py) package and
MeCab to do parsing.

Includes classes:

- JapaneseParser

"""

from io import StringIO
import sys
import os
import re
from typing import List
from natto import MeCab
import jaconv
from lute.parse.base import ParsedToken, AbstractParser
from lute.models.setting import UserSetting


class JapaneseParser(AbstractParser):
    """
    Japanese parser.

    This is only supported if mecab is installed.

    The parser uses natto-py library, and so should
    be able to find mecab automatically; if it can't,
    you may need to set the MECAB_PATH env variable,
    managed by UserSetting.set_value("mecab_path", p)
    """

    _is_supported = None
    _old_mecab_path = None

    @classmethod
    def is_supported(cls):
        """
        True if a natto MeCab can be instantiated,
        otherwise false.
        """
        mecab_path = os.environ.get("MECAB_PATH", "<NOTSET>")
        if (
            mecab_path == JapaneseParser._old_mecab_path
        ) and JapaneseParser._is_supported is not None:
            return JapaneseParser._is_supported

        b = False

        # Calling MeCab() prints to stderr even if the
        # exception is caught.  Suppress that output noise.
        temp_err = StringIO()
        try:
            sys.stderr = temp_err
            MeCab()
            b = True
        except:  # pylint: disable=bare-except
            b = False
        finally:
            sys.stderr = sys.__stderr__

        JapaneseParser._old_mecab_path = mecab_path
        JapaneseParser._is_supported = b
        return b

    @classmethod
    def name(cls):
        return "Japanese"

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        "Parse the string using MeCab."
        text = re.sub(r"[ \t]+", " ", text).strip()

        lines = []

        # If the string contains a "\n", MeCab appears to silently
        # remove it.  Splitting it works (ref test_JapaneseParser).
        # Flags: ref https://github.com/buruzaemon/natto-py:
        #    -F = node format
        #    -U = unknown format
        #    -E = EOP format
        with MeCab(r"-F %m\t%t\t%h\n -U %m\t%t\t%h\n -E EOP\t3\t7\n") as nm:
            for para in text.split("\n"):
                for n in nm.parse(para, as_nodes=True):
                    lines.append(n.feature)

        lines = [
            n.strip().split("\t") for n in lines if n is not None and n.strip() != ""
        ]

        # Production bug: JP parsing with MeCab would sometimes return a line
        # "0\t4" before an end-of-paragraph "EOP\t3\t7", reasons unknown.  These
        # "0\t4" tokens don't have any function, and cause problems in subsequent
        # steps of the processing in line_to_token(), so just remove them.
        lines = [n for n in lines if len(n) == 3]

        def line_to_token(lin):
            "Convert parsed line to a ParsedToken."
            term, node_type, third = lin
            is_eos = term in language.regexp_split_sentences
            if term == "EOP" and third == "7":
                term = "¶"
            is_word = node_type in "2678"
            return ParsedToken(term, is_word, is_eos or term == "¶")

        tokens = [line_to_token(lin) for lin in lines]
        return tokens

    # Hiragana is Unicode code block U+3040 - U+309F
    # ref https://stackoverflow.com/questions/72016049/
    #   how-to-check-if-text-is-japanese-hiragana-in-python
    def _char_is_hiragana(self, c) -> bool:
        return "\u3040" <= c <= "\u309F"

    def _string_is_hiragana(self, s: str) -> bool:
        return all(self._char_is_hiragana(c) for c in s)

    def get_reading(self, text: str):
        """
        Get the pronunciation for the given text.

        Returns None if the text is all hiragana, or the pronunciation
        doesn't add value (same as text).
        """
        if self._string_is_hiragana(text):
            return None

        flags = r"-O yomi"
        readings = []
        with MeCab(flags) as nm:
            for n in nm.parse(text, as_nodes=True):
                readings.append(n.feature)
        readings = [r.strip() for r in readings if r is not None and r.strip() != ""]

        ret = "".join(readings).strip()
        if ret in ("", text):
            return None

        jp_reading_setting = UserSetting.get_value("japanese_reading")
        if jp_reading_setting == "katakana":
            return ret
        if jp_reading_setting == "hiragana":
            return jaconv.kata2hira(ret)
        if jp_reading_setting == "alphabet":
            return jaconv.kata2alphabet(ret)
        raise RuntimeError(f"Bad reading type {jp_reading_setting}")



"""
Language mapping tests.
"""

from lute.models.language import Language
from lute.db import db
from tests.dbasserts import assert_sql_result
from tests.utils import make_text, add_terms


def test_save_new(empty_db):
    """
    Check language mappings and defaults.
    """
    sql = """select LgName, LgRightToLeft,
    LgShowRomanization, LgRegexpSplitSentences
    from languages"""
    assert_sql_result(sql, [], "empty table")

    lang = Language()
    lang.name = "abc"
    lang.dict_1_uri = "something"

    db.session.add(lang)
    db.session.commit()
    assert_sql_result(sql, ["abc; 0; 0; .!?"], "have language, defaults as expected")

    lang.right_to_left = True

    db.session.add(lang)
    db.session.commit()
    assert_sql_result(sql, ["abc; 1; 0; .!?"], "rtl is True")

    retrieved = db.session.query(Language).filter(Language.name == "abc").first()
    # print(retrieved)
    assert retrieved.name == "abc"
    assert retrieved.right_to_left is True, "retrieved is RTL"
    assert retrieved.show_romanization is False, "retrieved no roman"


def test_delete_language_removes_book_and_terms(app_context, spanish):
    """
    Test HACKY Language.delete() method to ensure deletes cascade.
    """
    add_terms(spanish, ["gato", "perro"])
    t = make_text("hola", "Hola amigo", spanish)
    db.session.add(t)
    db.session.commit()

    sqlterms = "select WoText from words order by WoText"
    sqlbook = "select BkTitle from books where BkTitle = 'hola'"
    assert_sql_result(sqlterms, ["gato", "perro"], "initial terms")
    assert_sql_result(sqlbook, ["hola"], "initial book")

    Language.delete(spanish)

    assert_sql_result(sqlbook, [], "book deleted")
    assert_sql_result(sqlterms, [], "terms deleted")



"""
Getting and saving bing image search results.
"""

import os
import re
import urllib.request
from flask import Blueprint, request, render_template, jsonify, current_app


bp = Blueprint("bing", __name__, url_prefix="/bing")


@bp.route("/search/<int:langid>/<string:text>/<string:searchstring>", methods=["GET"])
def bing_search(langid, text, searchstring):
    "Do an image search."

    # Searching for images slows acceptance tests.  If NO_BING_IMAGES
    # environment setting, don't do a search.
    if "NO_BING_IMAGES" in os.environ:
        return render_template(
            "imagesearch/index.html", langid=langid, text=text, images=[]
        )

    # dump("searching for " + text + " in " + language.getLgName())
    search = urllib.parse.quote(text)
    searchparams = searchstring.replace("###", search)
    url = "https://www.bing.com/images/search?" + searchparams
    content = ""
    with urllib.request.urlopen(url) as s:
        content = s.read().decode("utf-8")

    # Samples
    # <img class="mimg vimgld" ... data-src="https:// ...">
    # or
    # <img class="mimg rms_img" ... src="https://tse4.mm.bing ..." >

    pattern = r"(<img .*?>)"
    matches = re.findall(pattern, content, re.I)

    images = list(matches)

    def is_search_img(img):
        return not ('src="/' in img) and ("rms_img" in img or "vimgld" in img)

    def fix_data_src(img):
        return img.replace("data-src=", "src=")

    images = [fix_data_src(i) for i in images if is_search_img(i)]

    # Reduce image load count so we don't kill subpage loading.
    images = images[:25]

    def build_struct(image):
        src = "missing"
        m = re.search(r'src="(.*?)"', image)
        if m:
            src = m.group(1)
        return {"html": image, "src": src}

    data = [build_struct(i) for i in images]

    return render_template(
        "imagesearch/index.html", langid=langid, text=text, images=data
    )


def make_filename(text):
    return re.sub(r"\s+", "_", text) + ".jpeg"


@bp.route("/save", methods=["POST"])
def bing_save():
    """
    Save the posted image data to DATAPATH/userimages,
    returning the filename.
    """
    src = request.form["src"]
    text = request.form["text"]
    langid = request.form["langid"]

    datapath = current_app.config["DATAPATH"]
    imgdir = os.path.join(datapath, "userimages", langid)
    if not os.path.exists(imgdir):
        os.makedirs(imgdir)
    filename = make_filename(text)
    destfile = os.path.join(imgdir, filename)
    with urllib.request.urlopen(src) as response, open(destfile, "wb") as out_file:
        out_file.write(response.read())

    # This is the format of legacy Lute v2 data.
    image_url = f"/userimages/{langid}/{filename}"
    return jsonify({"filename": image_url})



"""
Term parent mapping.
"""

import csv
from sqlalchemy import text
from lute.db import db
from lute.models.term import Term, Status
from lute.term.model import Repository


## Exports


def export_terms_without_parents(language, outfile):
    "Export terms without parents in the language to filename outfile."
    # All existing terms that don't have parents.
    sig = Status.IGNORED
    sql = f"""
        SELECT w.WoTextLC
        FROM words w
        LEFT JOIN wordparents ON WpWoID = w.WoID
        WHERE w.WoLgID = {language.id}
          AND WpWoID IS NULL
          AND w.WoTokenCount = 1
          AND w.WoStatus != {sig}
    """
    data = db.session.execute(text(sql)).fetchall()
    terms = [term[0] for term in data]
    with open(outfile, "w", encoding="utf-8") as f:
        f.write("\n".join(terms))


def export_unknown_terms(book, outfile):
    "Export unknown terms in the book to outfile."
    lang = book.language
    unique_tokens = {
        t for txt in book.texts for t in lang.get_parsed_tokens(txt.text) if t.is_word
    }
    unique_lcase_toks = {lang.get_lowercase(t.token) for t in unique_tokens}

    lgid = lang.id
    known_terms_lc = (
        db.session.query(Term.text_lc)
        .filter(Term.language_id == lgid, Term.token_count == 1)
        .all()
    )
    known_terms_lc = [word[0] for word in known_terms_lc]

    newtoks = [t for t in unique_lcase_toks if t not in known_terms_lc]
    with open(outfile, "w", encoding="utf-8") as f:
        f.write("\n".join(newtoks))


## Imports


class BadImportFileError(Exception):
    """
    Raised if the import file is bad.
    """


def import_file(language, filename):
    """
    Validate and import file.

    Throws BadImportFileError if file contains invalid data.
    """
    import_data = _load_import_file(filename)
    _validate_data(import_data)
    return _do_import(language, import_data)


def _load_import_file(filename, encoding="utf-8-sig"):
    "Create array of hashes from file."
    importdata = []
    with open(filename, "r", encoding=encoding) as f:
        reader = csv.DictReader(f)

        fieldnames = reader.fieldnames
        if fieldnames is None:
            raise BadImportFileError("No mappings in file")
        _validate_data_fields(fieldnames)

        for line in reader:
            importdata.append(line)

    if len(importdata) == 0:
        raise BadImportFileError("No mappings in file")

    return importdata


def _validate_data_fields(field_list):
    "Check the keys in the file."
    for k in ["parent", "term"]:
        if k not in field_list:
            msg = "File must contain headings 'parent' and 'term'"
            raise BadImportFileError(msg)


def _validate_data(import_data):
    "All records must have parent, term."
    blanks = [
        hsh
        for hsh in import_data
        if hsh["term"].strip() == "" or hsh["parent"].strip() == ""
    ]
    if len(blanks) > 0:
        raise BadImportFileError("Term is required")


class ImportRecord:
    "Record in the import file."

    repo = None
    language = None

    @classmethod
    def set_context(cls, repo, language):
        "ImportRecord needs context for lookups."
        cls.repo = repo
        cls.language = language

    def _find(self, t):
        return ImportRecord.repo.find(ImportRecord.language.id, t)

    def __init__(self, hsh):
        self.ptext = hsh["parent"]
        self.parent = self._find(self.ptext)
        self.ctext = hsh["term"]
        self.child = self._find(self.ctext)

    @staticmethod
    def records(import_data):
        """
        Convert import data to records.

        This is called periodically during the import
        as each step updates the database.
        """
        return [ImportRecord(hsh) for hsh in import_data]


def _do_import(language, import_data):
    """
    Import records.
    """
    repo = Repository(db)
    ImportRecord.set_context(repo, language)

    updated = 0
    created = 0

    created, updated = _import_child_exists_parent_no(
        import_data, language, repo, created, updated
    )
    created, updated = _import_parent_exists_child_no(
        import_data, language, repo, created, updated
    )
    created, updated = _import_add_extra_parent_child_links(
        import_data, repo, created, updated
    )

    stats = {"created": created, "updated": updated}

    return stats


def _import_child_exists_parent_no(import_data, language, repo, created, updated):
    "Add parent and relationship."
    records = [
        p
        for p in ImportRecord.records(import_data)
        if p.parent is None and p.child is not None
    ]

    def _get_flash_msg(ptext):
        "Build a flash message for a new parent."
        all_children = [d.ctext for d in records if d.ptext == ptext]
        msg = f'Auto-created parent for "{all_children[0]}"'
        remaining = len(all_children) - 1
        if remaining > 0:
            msg += f" + {remaining} more"
        return msg

    # First add all the unique parents.
    ptexts = list({p.ptext for p in records})
    for p in ptexts:
        parent = repo.find_or_new(language.id, p)
        parent.flash_message = _get_flash_msg(p)
        repo.add(parent)
        created += 1
    repo.commit()

    # Then add all the relationships.
    for p in records:
        p.child.parents.append(p.ptext)
        repo.add(p.child)
        updated += 1
    repo.commit()

    return created, updated


def _import_parent_exists_child_no(import_data, language, repo, created, updated):
    "Add child and relationship."
    records = [
        p
        for p in ImportRecord.records(import_data)
        if p.parent is not None and p.child is None
    ]
    # Add all the children and relationships.
    for p in records:
        child = repo.find_or_new(language.id, p.ctext)
        if child.id is None:
            created += 1
        flash_msg = f'Auto-created and mapped to parent "{p.ptext}"'
        child.flash_message = flash_msg
        child.parents.append(p.ptext)
        repo.add(child)
    repo.commit()

    return created, updated


def _import_add_extra_parent_child_links(import_data, repo, created, updated):
    "Add parent to child if needed."
    records = [
        p
        for p in ImportRecord.records(import_data)
        if p.parent is not None
        and p.child is not None
        and p.parent.id != p.child.id
        and p.parent.text not in p.child.parents
    ]
    for p in records:
        p.child.parents.append(p.parent.text)
        repo.add(p.child)
        updated += 1
    repo.commit()

    return created, updated



"""
Lute test client.

Convenience wrapper around the browser.

IMPORTANT: on start, clears db, and disables backup.

Adds the home url to visited urls,
provides common functions like make_book.

Also calls the dev_api to set state, etc.

This module is "registered" to pytest in ./__init__.py
to get nicer assertion details.
"""

import time
import requests
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains


class LuteTestClient:  # pylint: disable=too-many-public-methods
    """
    The client!
    """

    ################################3
    # Setup

    def __init__(self, b, home):
        self.browser = b
        self.home = home
        self.visit("dev_api/wipe_db")
        self.visit("dev_api/load_demo_languages")
        self.visit("dev_api/disable_backup")

        response = requests.get(f"{home}/dev_api/language_ids", timeout=5)
        self.language_ids = response.json()

        self.start = time.perf_counter()
        self.last_step = self.start

    def load_demo_stories(self):
        "Load the demo stories."
        self.visit("dev_api/load_demo_stories")
        self.visit("/")
        self.clear_book_filter()

    def change_parser_registry_key(self, key, replacement):
        """
        Change a parser registry key to a replacement,
        effectively disabling that parser.
        """
        self.visit(f"dev_api/disable_parser/{key}/{replacement}")

    ################################3
    # Browsing

    def visit(self, suburl):
        "Visit a sub url under the base."
        if suburl.startswith("/"):
            suburl = suburl[1:]
        url = f"{self.home}/{suburl}"
        # print(f'visiting: {url}')
        self.browser.visit(url)

    def index(self):
        "Go to home page."
        self.browser.visit("")

    def clear_book_filter(self):
        "Clear all state.  Normally state is saved."
        self.browser.execute_script("clear_datatable_state()")
        time.sleep(0.1)

    def click_link(self, linktext):
        self.browser.links.find_by_text(linktext).click()

    ################################3
    # Languages

    def edit_language(self, langname, updates=None):
        """
        Edit a language.
        """
        self.visit("/")
        self.browser.find_by_css("#menu_settings").mouse_over()
        self.browser.find_by_id("lang_index").first.click()
        # WEIRD: find_by_text(langname) doesn't work ...
        self.browser.links.find_by_partial_text(langname).click()
        assert f"Edit {langname}" in self.browser.html
        updates = updates or {}
        for k, v in updates.items():
            if k == "exceptions_split_sentences":
                self.browser.find_by_css(f"#{k}").fill(v)
            else:
                raise RuntimeError(f"unhandled key {k}")
        self.browser.find_by_css("#submit").first.click()

    ################################3
    # Books

    def make_book(self, title, text, langname):
        "Create a book with title, text, and languagename."
        self.visit("book/new")
        self.browser.fill("text", text)
        self.browser.find_by_css("#title").fill(title)
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def make_book_from_file(self, title, filename, langname):
        "Create a book with title, content from filename, and languagename."
        self.visit("book/new")
        self.browser.attach_file("textfile", filename)
        self.browser.find_by_css("#title").fill(title)
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def make_book_from_url(self, url, langname):
        "Create a book with title, content from url, and languagename."
        self.visit("book/import_webpage")
        self.browser.find_by_css("#importurl").fill(url)
        self.browser.find_by_css("#import").first.click()
        time.sleep(0.1)  # hack
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def get_book_table_content(self):
        "Get book table content."
        css = "#booktable tbody tr"

        def _to_string(row):
            tds = row.find_by_css("td")
            rowtext = [td.text.strip() for td in tds]
            return "; ".join(rowtext).strip()

        rows = list(self.browser.find_by_css(css))
        return "\n".join([_to_string(row) for row in rows])

    ################################
    # Terms

    def _fill_term_form(self, b, updates):
        "Fill in the term form."
        for k, v in updates.items():
            if k == "language_id":
                b.select("language_id", v)
            elif k == "status":
                # This line didn't work:
                # iframe.choose('status', updates['status'])
                s = updates["status"]
                xp = "".join(
                    [
                        "//input[@type='radio'][@name='status']",
                        f"[@value='{s}']",
                        "/following-sibling::label",
                    ]
                )
                labels = b.find_by_xpath(xp)
                assert len(labels) == 1, "have matching radio button"
                label = labels[0]
                label.click()
            elif k in ("translation", "text"):
                b.find_by_css(f"#{k}").fill(v)
            elif k == "parents":
                for p in updates["parents"]:
                    xpath = [
                        # input w/ id
                        '//input[@id="parentslist"]',
                        # <tags> before it.
                        "/preceding-sibling::tags",
                        # <span> within the <tags> with class.
                        '/span[@class="tagify__input"]',
                    ]
                    span = b.find_by_xpath("".join(xpath))
                    span.type(p, slowly=False)
                    span.type(Keys.RETURN)
                    time.sleep(0.3)  # seconds
            elif k == "sync_status":
                if v:
                    b.check("sync_status")
                else:
                    b.uncheck("sync_status")
            else:
                raise RuntimeError(f"unhandled key {k}")

    def make_term(self, lang, updates):
        "Create a new term."
        self.visit("/")
        self.browser.find_by_css("#menu_terms").mouse_over()
        self.browser.find_by_id("term_index").first.click()
        self.browser.find_by_css("#term_actions").mouse_over()
        self.click_link("Create new")
        assert "New Term" in self.browser.html

        updates["language_id"] = self.language_ids[lang]
        b = self.browser
        self._fill_term_form(b, updates)
        b.find_by_css("#submit").first.click()

    def get_term_table_content(self):
        "Get term table content."
        self.visit("/")
        self.browser.find_by_css("#menu_terms").mouse_over()
        self.browser.find_by_id("term_index").first.click()
        css = "#termtable tbody tr"

        def _to_string(row):
            tds = row.find_by_css("td")
            rowtext = [td.text.strip() for td in tds]
            return "; ".join(rowtext).strip()

        rows = list(self.browser.find_by_css(css))
        return "\n".join([_to_string(row) for row in rows])

    ################################3
    # Reading/rendering

    def displayed_text(self):
        "Return the TextItems, with '/' at token boundaries."
        elements = self.browser.find_by_xpath('//span[contains(@class, "textitem")]')

        def _to_string(t):
            "Create string for token, eg 'cat (2)'."
            status = [
                c.replace("status", "")
                for c in t["class"].split(" ")
                if c.startswith("status") and c != "status0"
            ]
            if len(status) == 0:
                return t.text
            assert len(status) == 1, f"should only have 1 status on {t.text}"
            status = status[0]
            return f"{t.text} ({status})"

        etext = [_to_string(e) for e in elements]
        ret = "/".join(etext)
        if ret.endswith("/"):
            ret = ret[:-1]
        return ret

    ################################3
    # Reading, term actions

    def _get_element_for_word(self, word):
        "Helper, get the element."
        # print('getting ' + word)
        elements = self.browser.find_by_xpath('//span[contains(@class, "textitem")]')
        es = [e for e in elements if e.text == word]
        assert len(es) > 0, f"match for {word}"
        return es[0]

    def click_word(self, word):
        "Click a word in the reading frame."
        self._get_element_for_word(word).click()

    def shift_click_words(self, words):
        "Shift-click words."
        # https://stackoverflow.com/questions/27775759/
        #   send-keys-control-click-in-selenium-with-python-bindings
        # pylint: disable=protected-access
        els = [self._get_element_for_word(w)._element for w in words]
        ac = ActionChains(self.browser.driver).key_down(Keys.SHIFT)
        for e in els:
            ac = ac.click(e)
        ac = ac.key_up(Keys.SHIFT)
        ac.perform()

    def press_hotkey(self, hotkey):
        "Send a hotkey."
        el = self.browser.find_by_tag("body")
        map_to_js_keycode = {
            "1": 49,
            "2": 50,
            "3": 51,
            "4": 52,
            "5": 53,
            "i": 73,
            "w": 87,
            "c": 67,
            "t": 84,
            "m": 77,
            "h": 72,
            "up": 38,
            "down": 40,
        }
        jscode = map_to_js_keycode[hotkey.lower()]
        shift_pressed = "true" if hotkey in ["C", "T"] else "false"

        # This was the only way I could get this to work:
        script = f"""jQuery.event.trigger({{
          type: 'keydown',
          which: {jscode},
          shiftKey: '{shift_pressed}'
        }});"""
        # pylint: disable=protected-access
        self.browser.execute_script(script, el._element)
        time.sleep(0.2)  # Or it's too fast.
        # print(script)
        # Have to refresh the content to query the dom ...
        # Unfortunately, I can't see how to refresh without reloading
        self.browser.reload()

    def click_word_fill_form(self, word, updates=None):
        """
        Click a word in the reading frame, fill in the term form iframe.
        """
        self.click_word(word)
        updates = updates or {}

        should_refresh = False
        with self.browser.get_iframe("wordframe") as iframe:
            self._fill_term_form(iframe, updates)
            iframe.find_by_css("#submit").first.click()

            # Only refresh the reading frame if everything was ok.
            # Some submits will fail due to validation errors,
            # and we want to look at them.
            if "updated" in iframe.html:
                should_refresh = True

        # Have to refresh the content to query the dom ...
        # Unfortunately, I can't see how to refresh without reloading
        if should_refresh:
            self.browser.reload()

    ################################3
    # Misc.

    def elapsed(self, step):
        """
        Helper method for sorting out slowness.

        For the step, gives elapsed time since start of
        the LuteBrowser, and since the last recorded step.

        To see this data, run the acc. tests with '-s', eg:

        pytest tests/acceptance/test_smoke.py --port=5000 -s
        """
        now = time.perf_counter()
        since_start = now - self.start
        print(step)
        print(f"total elapsed: {since_start}")
        print(f"since last:    {now - self.last_step}")
        self.last_step = now

    def sleep(self, seconds):
        "Nap."
        time.sleep(seconds)

    def get_temp_file_content(self, filename):
        "Get book table content."
        response = requests.get(
            f"{self.home}/dev_api/temp_file_content/{filename}", timeout=1
        )
        return response.text



"""
Term business object and repository.

Terms are converted to and from lute.models.term.Term objects to save
them in the database.
"""

import re
import functools
from sqlalchemy import and_, text as sqlalchtext

from lute.models.term import Term as DBTerm, TermTag
from lute.models.language import Language


class Term:  # pylint: disable=too-many-instance-attributes
    """
    Term business object.  All class members are primitives.
    """

    def __init__(self):
        # The ID of the DBTerm.
        self.id = None
        # A language object is required as the Term bus. object
        # must downcase the text and the original_text to see
        # if anything has changed.
        self._language = None
        # Ideally this wouldn't be needed, but the term form
        # populates this field with the (primitive) language id.
        self.language_id = None
        # The text.
        self.text = None
        self.text_lc = None
        # The original text given to the DTO, to track changes.
        self.original_text = None
        self.status = 1
        self.translation = None
        self.romanization = None
        self.sync_status = False
        self.term_tags = []
        self.flash_message = None
        self.parents = []
        self.current_image = None

    def __repr__(self):
        return (
            f'<Term BO "{self.text}" lang_id={self.language_id} lang={self.language}>'
        )

    @property
    def language(self):
        "Use or get the language."
        if self._language is not None:
            return self._language
        return Language.find(self.language_id)

    @language.setter
    def language(self, lang):
        if not isinstance(lang, Language):
            raise ValueError("not a language")
        self._language = lang

    def text_has_changed(self):
        "Check the downcased original text with the current text."
        # print(f'checking if changed, orig = "{self.original_text}", text = "{self.text}"')
        if self.original_text in ("", None):
            return False

        def get_lc(s):
            return self.language.get_lowercase(s)

        return get_lc(self.original_text) != get_lc(self.text)


class TermReference:
    "Where a Term has been used in books."

    def __init__(
        self, bookid, txid, pgnum, title, sentence=None
    ):  # pylint: disable=too-many-arguments
        self.book_id = bookid
        self.text_id = txid
        self.page_number = pgnum
        self.title = title
        self.sentence = sentence


class Repository:
    """
    Maps Term BO to and from lute.model.Term.
    """

    def __init__(self, _db):
        self.db = _db

        # Identity map for business lookup.
        # Note that the same term is stored
        # a few times: by upper case or lower case text,
        # plus language ID.
        self.identity_map = {}

    def _id_map_key(self, langid, text):
        return f"key-{langid}-{text}"

    def _add_to_identity_map(self, t):
        "Add found term to identity map."
        for txt in t.text, t.text_lc:
            k = self._id_map_key(t.language_id, txt)
            self.identity_map[k] = t

    def _search_identity_map(self, langid, txt):
        "Search the map w/ the given bus. key."
        k = self._id_map_key(langid, txt)
        if k in self.identity_map:
            return self.identity_map[k]
        return None

    def load(self, term_id):
        "Loads a Term business object for the DBTerm with the id."
        dbt = DBTerm.find(term_id)
        if dbt is None:
            raise ValueError(f"No term with id {term_id} found")
        term = self._build_business_term(dbt)
        self._add_to_identity_map(term)
        return term

    def find(self, langid, text):
        """
        Return a Term business object for the DBTerm with the langid and text.
        If no match, return None.
        """
        term = self._search_identity_map(langid, text)
        if term is not None:
            return term

        spec = self._search_spec_term(langid, text)
        dbt = DBTerm.find_by_spec(spec)
        if dbt is None:
            return None
        term = self._build_business_term(dbt)
        self._add_to_identity_map(term)
        return term

    def find_or_new(self, langid, text):
        """
        Return a Term business object for the DBTerm with the langid and text.
        If no match, return a new term with the text and language.

        If it's new, don't add to the identity map ... it's not saved yet,
        and so if we search for it again we should hit the db again.
        """
        t = self.find(langid, text)
        if t is not None:
            return t

        spec = self._search_spec_term(langid, text)
        t = Term()
        t.language = spec.language
        t.language_id = langid
        t.text = text
        t.text_lc = spec.text_lc
        t.romanization = spec.language.parser.get_reading(text)
        t.original_text = text

        return t

    def find_matches(self, langid, text, max_results=50):
        """
        Return array of Term business objects for the DBTerms
        with the same langid, matching the text.
        If no match, return [].
        """
        spec = self._search_spec_term(langid, text)
        text_lc = spec.text_lc
        search = text_lc.strip() if text_lc else ""
        if search == "":
            return []

        matches = (
            self.db.session.query(DBTerm)
            .filter(
                and_(
                    DBTerm.language_id == langid,
                    DBTerm.text_lc.like("%" + search + "%"),
                )
            )
            .all()
        )

        exact = [t for t in matches if t.text_lc == text_lc]
        remaining = [t for t in matches if t.text_lc != text_lc]

        # Split remaining into things that start with the search term
        # and the rest.
        def partition(arr, pred):
            meets = []
            notmeets = []
            for x in arr:
                if pred(x):
                    meets.append(x)
                else:
                    notmeets.append(x)
            return (meets, notmeets)

        remain_starts_with, remain_rest = partition(
            remaining, lambda t: t.text_lc.startswith(search)
        )

        # Sort terms with children to top,
        # then alphabetically.
        def compare(item1, item2):
            # More children sort to top.
            c1 = len(item1.children)
            c2 = len(item2.children)
            if c1 > c2:
                return -1
            if c1 < c2:
                return 1

            # Equal children sort alphabetically.
            t1 = item1.text_lc
            t2 = item2.text_lc
            if t1 < t2:
                return -1
            if t1 > t2:
                return 1

            # Failsafe, should never get here.
            return 0

        remain_starts_with.sort(key=functools.cmp_to_key(compare))
        remain_rest.sort(key=functools.cmp_to_key(compare))
        ret = exact + remain_starts_with + remain_rest
        ret = ret[:max_results]
        return [self._build_business_term(t) for t in ret]

    def get_term_tags(self):
        "Get all available term tags, helper method."
        tags = self.db.session.query(TermTag).all()
        return [t.text for t in tags]

    def add(self, term):
        """
        Add a term to be saved to the db session.
        Returns DB Term for tests and verification only,
        clients should not change it.
        """
        dbterm = self._build_db_term(term)
        self.db.session.add(dbterm)
        return dbterm

    def delete(self, term):
        """
        Add term to be deleted to session.
        """
        spec = self._search_spec_term(term.language_id, term.text)
        dbt = DBTerm.find_by_spec(spec)
        if dbt is None:
            return
        self.db.session.delete(dbt)

    def commit(self):
        """
        Commit everything.
        """
        self.db.session.commit()

    def _search_spec_term(self, langid, text):
        """
        Make a term to get the correct text_lc to search for.

        Creating a term does parsing and correct downcasing,
        so term.language.id and term.text_lc match what the
        db would contain.
        """
        lang = Language.find(langid)
        return DBTerm(lang, text)

    def _build_db_term(self, term):
        "Convert a term business object to a DBTerm."
        if term.text is None:
            raise ValueError("Text not set for term")

        spec = self._search_spec_term(term.language_id, term.text)
        t = DBTerm.find_by_spec(spec)
        if t is None:
            t = DBTerm()

        t.language = spec.language
        t.text = term.text
        t.original_text = term.text
        t.status = term.status
        t.translation = term.translation
        t.romanization = term.romanization
        t.sync_status = term.sync_status
        t.set_current_image(term.current_image)

        if term.flash_message is not None:
            t.set_flash_message(term.flash_message)
        else:
            t.pop_flash_message()

        termtags = []
        for s in term.term_tags:
            termtags.append(TermTag.find_or_create_by_text(s))
        t.remove_all_term_tags()
        for tt in termtags:
            t.add_term_tag(tt)

        termparents = []
        lang = spec.language
        create_parents = [
            p
            for p in term.parents
            if p is not None
            and p != ""
            and lang.get_lowercase(term.text) != lang.get_lowercase(p)
        ]
        # print('creating parents: ' + ', '.join(create_parents))
        for p in create_parents:
            termparents.append(self._find_or_create_parent(p, lang, term, termtags))
        t.remove_all_parents()
        for tp in termparents:
            t.add_parent(tp)

        if len(termparents) != 1:
            t.sync_status = False

        return t

    def _find_or_create_parent(self, pt, language, term, termtags) -> DBTerm:
        spec = self._search_spec_term(language.id, pt)
        p = DBTerm.find_by_spec(spec)

        if p is not None:
            if (p.translation or "") == "":
                p.translation = term.translation
            if (p.get_current_image() or "") == "":
                p.set_current_image(term.current_image)
            return p

        p = DBTerm(language, pt)
        p.status = term.status
        p.translation = term.translation
        p.set_current_image(term.current_image)
        for tt in termtags:
            p.add_term_tag(tt)

        return p

    def _build_business_term(self, dbterm):
        "Create a Term bus. object from a lute.model.term.Term."
        term = Term()
        term.id = dbterm.id
        term.language = dbterm.language
        term.language_id = dbterm.language.id

        # Remove zero-width spaces (zws) from strings for user forms.
        text = dbterm.text
        zws = "\u200B"  # zero-width space
        text = text.replace(zws, "")
        term.text_lc = dbterm.text_lc
        term.original_text = text
        term.text = text

        term.status = dbterm.status
        term.translation = dbterm.translation
        term.romanization = dbterm.romanization
        term.sync_status = dbterm.sync_status
        term.current_image = dbterm.get_current_image()
        term.flash_message = dbterm.get_flash_message()
        term.parents = [p.text for p in dbterm.parents]
        term.romanization = dbterm.romanization
        term.term_tags = [tt.text for tt in dbterm.term_tags]

        return term

    ## References.

    def find_references(self, term):
        """
        Return references of term, children, and parents.
        """
        spec = self._search_spec_term(term.language_id, term.text)
        searchterm = DBTerm.find_by_spec(spec)
        if searchterm is None:
            searchterm = spec

        references = {
            "term": self._get_references(searchterm),
            "children": self._get_child_references(searchterm),
            "parents": self._get_parent_references(searchterm),
        }
        return references

    def _build_term_references(self, term_lc, rows):
        ret = []
        zws = chr(0x200B)  # zero-width space
        for row in rows:
            sentence = row[4].strip()
            pattern = f"{zws}({term_lc}){zws}"

            def replace_match(m):
                return f"{zws}<b>{m.group(0)}</b>{zws}"

            sentence = re.sub(pattern, replace_match, sentence, flags=re.IGNORECASE)
            sentence = sentence.replace(zws, "").replace("¶", "")
            ret.append(TermReference(row[0], row[1], row[2], row[3], sentence))
        return ret

    def _get_references(self, term):
        if term is None:
            return []

        term_lc = term.text_lc
        query = sqlalchtext(
            f"""
            SELECT DISTINCT
                texts.TxBkID,
                TxID,
                TxOrder,
                BkTitle || ' (' || TxOrder || '/' || pc.c || ')' AS TxTitle,
                SeText
            FROM sentences
            INNER JOIN texts ON TxID = SeTxID
            INNER JOIN books ON BkID = texts.TxBkID
            INNER JOIN (
                SELECT TxBkID, COUNT(*) AS c
                FROM texts
                GROUP BY TxBkID
            ) pc ON pc.TxBkID = texts.TxBkID
            WHERE TxReadDate IS NOT NULL
            AND LOWER(SeText) LIKE :pattern
            AND BkLgID = {term.language.id}
            LIMIT 20
        """
        )

        pattern = f"%{chr(0x200B)}{term_lc}{chr(0x200B)}%"
        params = {"pattern": pattern}
        result = self.db.session.execute(query, params)
        return self._build_term_references(term_lc, result)

    def _get_all_refs(self, terms):
        all_references = []
        for term in terms:
            references = self._get_references(term)
            all_references.extend(references)
        return all_references

    def _get_parent_references(self, term):
        parent_references = []
        for parent in term.parents:
            parent_term_lc = parent.text_lc
            references = self._get_family_references(parent, term)
            parent_references.append({"term": parent_term_lc, "refs": references})
        return parent_references

    def _get_family_references(self, parent, term):
        if term is None or parent is None:
            return []
        family = [parent] + parent.children
        family = [t for t in family if t.id != term.id]
        return self._get_all_refs(family)

    def _get_child_references(self, term):
        if term is None:
            return []
        return self._get_all_refs(term.children)



"""Database assertions.

This module is "registered" to pytest in ./__init__.py
to get nicer assertion details.
"""

import os
from sqlalchemy import create_engine
from sqlalchemy.sql import text

from lute.config.app_config import AppConfig


def __sqlite_uri():
    """
    Get the application Sqlite URL from the config file.
    """
    this_files_dir = os.path.dirname(os.path.realpath(__file__))
    conf_file = os.path.normpath(
        os.path.join(this_files_dir, "..", "lute", "config", "config.yml")
    )
    ac = AppConfig(conf_file)
    return ac.sqliteconnstring


def assert_sql_result(sql, expected, msg=""):
    """
    Checks sql results (stringized) against expected.

    e.g. with sql = "select 1, 2", expected would be [ "1; 2" ]

    This replaces all zero-width strings with '/'
    """
    uri = __sqlite_uri()
    engine = create_engine(uri)
    conn = engine.connect()
    result = conn.execute(text(sql))
    zws = "\u200B"  # zero-width space
    actual = ["; ".join([f"{s}".replace(zws, "/") for s in row]) for row in result]
    assert actual == expected, msg


def assert_record_count_equals(sql, expected, message=""):
    """
    Checks count of records of sql results (stringized) against expected.

    e.g. with sql = "select 1, 2", expected would be 1
    """

    uri = __sqlite_uri()
    engine = create_engine(uri)
    conn = engine.connect()

    # Check if the SQL query is a SELECT statement
    if not sql.lower().startswith("select"):
        sql = f"SELECT * FROM {sql}"

    query = text(sql)
    result = conn.execute(query)

    actual = ["; ".join([f"{s}" for s in row]) for row in result]

    if len(actual) != expected:
        message += f" ... got data:\n\n{actual}\n"

    assert len(actual) == expected, message



"""
Db management.
"""

from sqlalchemy import text
from lute.db import db
from lute.models.setting import UserSetting


def delete_all_data():
    """
    DANGEROUS!  Delete everything, restore user settings, clear sys settings.

    NO CHECKS ARE PERFORMED.
    """

    # Setting the pragma first ensures cascade delete.
    statements = [
        "pragma foreign_keys = ON",
        "delete from languages",
        "delete from tags",
        "delete from tags2",
        "delete from settings",
    ]
    for s in statements:
        db.session.execute(text(s))
    db.session.commit()
    UserSetting.load()



"""
Theming service.

Themes are stored in the css folder.  Current theme is saved in
UserSetting.
"""

import os
from glob import glob
from lute.models.setting import UserSetting
from lute.db import db

default_entry = ("-", "(default)")


def _css_path():
    """
    Path to css in this folder.
    """
    thisdir = os.path.dirname(__file__)
    theme_dir = os.path.join(thisdir, "css")
    return os.path.abspath(theme_dir)


def list_themes():
    """
    List of theme file names and user-readable name.
    """
    theme_glob = os.path.join(_css_path(), "*.css")

    def _make_display_name(s):
        ret = os.path.basename(s)
        ret = ret.replace(".css", "").replace("_", " ")
        return ret

    glob_results = glob(theme_glob)
    themes = [(os.path.basename(f), _make_display_name(f)) for f in glob_results]
    return [default_entry] + themes


def get_current_css():
    """
    Return the current css pointed at by the current_theme user setting.
    """
    current_theme = UserSetting.get_value("current_theme")
    if current_theme == default_entry[0]:
        return ""
    theme_css = os.path.join(_css_path(), current_theme)
    if not os.path.exists(theme_css):
        # _Probably_ should raise an error here ...
        return ""
    with open(theme_css, "r", encoding="utf-8") as f:
        return f.read()


def next_theme():
    """
    Move to the next theme in the list of themes.
    """
    current_theme = UserSetting.get_value("current_theme")
    themes = [t[0] for t in list_themes()]
    themes.append(default_entry[0])
    for i in range(0, len(themes)):  # pylint: disable=consider-using-enumerate
        if themes[i] == current_theme:
            new_index = i + 1
            break
    UserSetting.set_value("current_theme", themes[new_index])
    db.session.commit()
